import pygame
from mlgame.game.paia_game import GameResultState, GameStatus
from mlgame.view.view_model import create_polygon_view_data
from .env import *


class Point(pygame.sprite.Sprite):
    def __init__(self, game, coordinate):
        self.group = game.all_points
        pygame.sprite.Sprite.__init__(self, self.group)
        self.game = game
        # self.x, self.y = (coordinate[0] - TILESIZE / PPM, -coordinate[1] + TILESIZE / PPM)
        self.x, self.y = (coordinate[0], coordinate[1])

    def get_info(self):
        return {
            "coordinate": (self.x, self.y)
        }

    def get_progress_data(self):
        asset_data = {}
        return asset_data


class End_point(Point):
    def __init__(self, game, coordinate):
        Point.__init__(self, game, coordinate)
        self.rect = pygame.Rect(self.x, self.y, TILESIZE * 3, TILESIZE * 3)

    def update(self, *args, **kwargs) -> None:
        self.detect_cars_collision()

    def detect_cars_collision(self):
        hits = pygame.sprite.spritecollide(self, self.game.cars, False)
        for hit in hits:
            if hit.is_running:
                hit.check_point += 1
                hit.end_frame = self.game.frame
                hit.is_completed = True
                self.game.eliminated_user.append(hit)  # TODO #外部注入
                if self.game.user_num==1:
                    self.game.state = GameResultState.PASSED
                elif self.game.user_num>1:
                    self.game.state = GameResultState.FINISH
                hit.is_running = False
                hit.status = GameStatus.GAME_PASS
    def get_info(self):
        return {
            "coordinate": self.rect.center
        }
    def get_progress_data(self):
        asset_data = {"type": "image",
                      "x": self.rect.x,
                      "y": self.rect.y,
                      "width": self.rect.width,
                      "height": self.rect.height,
                      "image_id": "target",
                      "angle": 0}
        return asset_data


class Check_point(Point):
    def __init__(self, game, vertices):
        Point.__init__(self, game, vertices[0])
        self.rect = pygame.Rect(vertices[0][0], vertices[0][1], abs(vertices[2][0] - vertices[1][0]), abs(vertices[2][1]-vertices[1][1]))
        self.car_has_hit = []
        self.vertices = vertices
        self.color = YELLOW

    def update(self, *args, **kwargs) -> None:
        self.color = YELLOW
        self.detect_cars_collision()

    def detect_cars_collision(self):
        # TODO 車子和檢查點的碰撞有點不精確
        hits = pygame.sprite.spritecollide(self, self.game.cars, False)
        for hit in hits:
            self.color = GREEN
            if hit.status and hit not in self.car_has_hit:
                hit.check_point += 1
                hit.end_frame = self.game.frame
                self.car_has_hit.append(hit)

    def get_progress_data(self):
        # asset_data = {"type": "rect",
        #               "name": 'check_point',
        #               "x": self.rect.x,
        #               "y": self.rect.y,
        #               "angle": 0,
        #               "width": self.rect.width,
        #               "height": self.rect.height,
        #               "color": YELLOW}
        asset_data = create_polygon_view_data("check_point", [list(v) for v in self.vertices], self.color)
        return asset_data
    def get_info(self):
        return {
            "coordinate": self.rect.center
        }

class Outside_point(Point):
    '''
    if car colliding these point, car will be swich to start point
    '''

    def __init__(self, game, coordinate):
        Point.__init__(self, game, coordinate)
        # self.image = pygame.Surface((TILESIZE, TILESIZE))
        # self.image.fill(BLUE)
        # self.rect = self.image.get_rect()
        self.rect = pygame.Rect(self.x, self.y, TILESIZE * 3, TILESIZE * 3)

    def update(self, *args, **kwargs) -> None:
        self.detect_cars_collision()

    def detect_cars_collision(self):
        hits = pygame.sprite.spritecollide(self, self.game.cars, False)
        for hit in hits:
            if hit.status:
                hit.body.position = (hit.x, hit.y)
                hit.body.linearVelocity = 0, 0
                pass



import json

import pygame
from .env import *

class Map:
    def __init__(self, filename):
        self.data = []
        with open(filename) as json_file:
            data = json.load(json_file)
            self.tileHeight = data["height"]
            self.tileWidth = data["width"]
            map = []
            for i in range(self.tileHeight):
                self.data.append(data["layers"][0]["data"][i * self.tileWidth:(i + 1) * self.tileWidth])
        # with open(filename,'rt') as f:
        #     for line in f:
        #         data = []
        #         line.strip() # ignore"\n"
        #         for c in line:
        #             data.append(c)
        #         self.data.append(data)
        # self.tileWidth = len(self.data[0])
        # self.tileHeight = len(self.data)
        self.width = self.tileWidth * TILESIZE
        self.height = self.tileHeight * TILESIZE

class Camera:
    def __init__(self, width, height):
        self.camera = pygame.Rect(0,0,width,height)
        self.width = width
        self.height = height

    def apply(self, entity):
        return entity.rect.move(self.camera.topleft)

    def update(self, target):
        x = -target.rect.centerx + int(WIDTH / 2)
        y = -target.rect.centery + int(HEIGHT / 2)

        # limit scrolling to map size
        x = min(0,x)
        x = max(-(self.width - WIDTH+1), x)
        y = min(0,y)
        y = max(-(self.height - HEIGHT), y)
        self.camera = pygame.Rect(x, y, self.width, self.height)


from os import path

from mlgame.view.view_model import create_asset_init_data, create_image_view_data
from TankBattleMode import TankBattleMode
from TankPlayer import TankPlayer
from env import MAP_DIR


class TestTankBattleMode(object):
    map_path = path.join(MAP_DIR, "map_00.tmx")
    mode = TankBattleMode(map_path, 1200, True)

    def test_calculate_score_type(self):
        assert type(self.mode.calculate_score()) == tuple
        assert type(self.mode.calculate_score()[0]) == int
        assert type(self.mode.calculate_score()[1]) == int

    def test_is_list_type_of_get_1P_command(self):
        assert type(self.mode.get_1P_command()) == list

    def test_is_list_type_of_get_2P_command(self):
        assert type(self.mode.get_2P_command()) == list

    def test_draw_sprite_data(self):
        sprite_data = self.mode.draw_sprite_data()
        image_init_data_template = create_image_view_data("1P", 0, 0, 60, 60, 0)
        for data in sprite_data:
            assert data.keys() == image_init_data_template.keys()
            for key in data.keys():
                assert type(data[key]) == type(image_init_data_template[key])


class TestTankPlayer(object):
    construction = {"_id": 1, "_no": 0, "x": 0, "y": 0, "width": 60, "height": 60}
    player = TankPlayer(construction)

    # TODO how to test update
    def test_update(self):
        pass

    # TODO how to test rotate
    def test_rotate(self):
        pass

    def get_shoot_info(self):
        return {"id": 0, "center_pos": (0, 0), "rot": 0}

    def test_get_player_shoot_info(self):
        shoot_info = self.player.create_shoot_info()
        assert shoot_info.keys() == self.get_shoot_info().keys()
        for key in shoot_info.keys():
            assert type(shoot_info[key]) == type(self.get_shoot_info()[key])

    def test_arg(self):
        assert self.player.origin_size == (60, 60)
        assert self.player.speed == 8
        assert type(self.player.angle) == int
        assert type(self.player.score) == int
        assert type(self.player.used_frame) == int
        assert type(self.player.rot) == int
        assert self.player.rot_speed == 45
        assert 100 >= self.player.shield > 0
        assert 10 >= self.player.power > 0
        assert 100 >= self.player.oil > 0
        assert type(self.player.last_turn_frame) == int
        assert type(self.player.is_shoot) == bool
        assert self.player.is_alive
        assert type(self.player.is_turn) == bool
        assert type(self.player.is_forward) == bool
        assert type(self.player.is_backward) == bool

    def test_move_arg(self):
        move = {"left_up": (-self.player.speed, -self.player.speed),
                "right_up": (self.player.speed, -self.player.speed),
                "left_down": (-self.player.speed, self.player.speed),
                "right_down": (self.player.speed, self.player.speed),
                "left": (-self.player.speed, 0), "right": (self.player.speed, 0), "up": (0, -self.player.speed),
                "down": (0, self.player.speed)}
        assert self.player.move_dict == move

    def test_act(self):
        # TODO 增加限制type
        # TODO how test shoot act
        self.player.act(["SHOOT"])
        # assert self.player.is_shoot and self.player.power == 9 and self.player.last_shoot_frame == 50
        self.player.act(["TURN_LEFT"])
        assert self.player.is_turn and round(self.player.oil, 1) == 99.9
        self.player.act(["TURN_RIGHT"])
        assert self.player.is_turn and round(self.player.oil, 1) == 99.8
        self.player.act(["FORWARD"])
        assert self.player.is_forward and not self.player.is_backward and round(self.player.oil, 1) == 99.7
        self.player.act(["BACKWARD"])
        assert self.player.is_backward and not self.player.is_forward and round(self.player.oil, 1) == 99.6

    def go_straight(self, forward_or_backward: str):
        self.player.rot = 0
        self.player.hit_rect.center = self.player.origin_center
        if forward_or_backward == "forward":
            self.player.forward()
        else:
            self.player.backward()
        return self.player.origin_center, self.player.hit_rect.center

    # TODO how better
    def test_go_straight(self):
        speed = self.player.speed
        origin_center, now_center = self.go_straight("forward")
        assert (origin_center[0]-speed, origin_center[1]) == now_center
        origin_center, now_center = self.go_straight("backward")
        assert (origin_center[0]+speed, origin_center[1]) == now_center

    def turn(self, right_or_left: str, is_turn: bool):
        origin_rot = self.player.rot
        self.player.is_turn = is_turn
        self.player.last_turn_frame = 500
        if right_or_left == "right":
            self.player.turn_right()
        else:
            self.player.turn_left()
        return self.player.last_turn_frame, origin_rot, self.player.rot, self.player.is_turn

    # TODO how better
    def test_turn_method(self):
        turn_cd, origin_rot, new_rot, is_turn = self.turn("right", True)
        assert turn_cd == 500 and new_rot == origin_rot and is_turn
        turn_cd, origin_rot, new_rot, is_turn = self.turn("right", False)
        assert turn_cd == 0 and new_rot == origin_rot-self.player.rot_speed and is_turn
        turn_cd, origin_rot, new_rot, is_turn = self.turn("left", True)
        assert turn_cd == 500 and new_rot == origin_rot and is_turn
        turn_cd, origin_rot, new_rot, is_turn = self.turn("left", False)
        assert turn_cd == 0 and new_rot == origin_rot+self.player.rot_speed and is_turn

    def collide_with_walls(self, is_forward: bool):
        self.player.rot = 0
        self.player.hit_rect.center = (100, 100)
        old_center = self.player.hit_rect.center
        self.player.is_forward = is_forward
        self.player.collide_with_walls()
        return old_center, self.player.hit_rect.center

    def test_collide_with_walls(self):
        speed = self.player.speed
        old_center, new_center = self.collide_with_walls(is_forward=True)
        assert (old_center[0]+speed, old_center[1]) == new_center
        old_center, new_center = self.collide_with_walls(is_forward=False)
        assert (old_center[0]-speed, old_center[1]) == new_center

    def collide_with_bullets(self, shield: int, lives: int):
        self.player.shield = shield
        self.player.lives = lives
        self.player.collide_with_bullets()
        return {"shield": self.player.shield, "lives": self.player.lives, "pos": self.player.rect.center}

    def test_collide_with_bullets(self):
        self.player.rect.center = (50, 50)
        self.collide_with_bullets(1, 1)
        assert self.player.shield == 100
        assert self.player.lives == 0
        assert self.player.rect.center == self.player.origin_center
        self.player.rect.center = (50, 50)
        self.collide_with_bullets(1, 2)
        assert self.player.shield == 100
        assert self.player.lives == 1
        assert self.player.rect.center == self.player.origin_center
        self.player.rect.center = (50, 50)
        self.collide_with_bullets(100, 2)
        assert 100 > self.player.shield > 89
        assert self.player.lives == 2
        assert self.player.rect.center == (50, 50)
        assert self.player.rect.center != self.player.origin_center

    def test_get_power(self):
        self.player.get_power(-10)
        assert self.player.power == 0
        self.player.get_power(10)
        assert self.player.power == 10
        self.player.get_power(-11)
        assert self.player.power == 0
        self.player.get_power(11)
        assert self.player.power == 10
        self.player.get_power(-5)
        assert self.player.power == 5
        self.player.get_power(3)
        assert self.player.power == 8

    def test_get_oil(self):
        self.player.get_oil(-100)
        assert self.player.oil == 0
        self.player.get_oil(100)
        assert self.player.oil == 100
        self.player.get_oil(-110)
        assert self.player.oil == 0
        self.player.get_oil(110)
        assert self.player.oil == 100
        self.player.get_oil(-50)
        assert self.player.oil == 50
        self.player.get_oil(30)
        assert self.player.oil == 80

    def get_info(self):
        return {"id": f"{0}P",
                "x": 0,
                "y": 0,
                "speed": 0,
                "score": 0,
                "power": 0,
                "oil": 0,
                "shield": 0,
                "lives": 0
                }

    def test_get_player_info(self):
        info = self.player.get_info()
        assert info.keys() == self.get_info().keys()
        for key in info.keys():
            assert type(info[key]) == type(self.get_info()[key])

    def get_result(self):
        return {"id": f"{0}P", "x": 0, "y": 0, "score": 0, "shield": 0, "lives": 0}

    def test_get_result(self):
        result = self.player.get_result()
        assert result.keys() == self.get_result().keys()
        for key in result.keys():
            assert type(result[key]) == type(self.get_result()[key])

    def get_image_data(self):
        return {ID: f"{0}P", X: 0, Y: 0, WIDTH: 0, HEIGHT: 0, ANGLE: 0}

    def test_get_image_data(self):
        image_data = self.player.get_image_data()
        assert image_data.keys() == self.get_image_data().keys()
        for key in image_data.keys():
            assert type(image_data[key]) == type(self.get_image_data()[key])

    def test_get_image_init_data(self):
        image_init_data = self.player.get_image_init_data()
        image_init_data_template = create_asset_init_data("1P", 60, 60, "path", "url")
        for data in image_init_data:
            assert data.keys() == image_init_data_template.keys()
            for key in data.keys():
                assert type(data[key]) == type(image_init_data_template[key])



import urllib3.util
from azure.storage.blob import ContainerClient, BlobClient
from loguru import logger


import json



def upload_data_to_azure_blob(az_path_blob_url: str, file_name: str, data: list) -> bool:
    """
    Uploads a dictionary to Azure Blob Storage.

    :param az_path_blob_url: The URL of the Azure container.
    :param file_name: The name of the blob to create.
    :param data: The dictionary to upload.
    :return: True if upload is successful, False otherwise.
    """
    # Create a ContainerClient
    path_blob_url = urllib3.util.parse_url(az_path_blob_url)
    file_blob_url = str(path_blob_url).replace(path_blob_url.path, path_blob_url.path+ "/" + file_name)

    # Create a blob client
    blob_client = BlobClient.from_blob_url(file_blob_url)

    # Convert the dictionary to a JSON string
    json_data = json.dumps(data)

    # Upload the JSON string to the blob
    blob_client.upload_blob(json_data, overwrite=True)




def upload_data_to_azure_blob_by_container(az_container_url: str, blob_name: str, data: list) -> bool:
    """
    Uploads a dictionary to Azure Blob Storage.

    :param az_container_url: The URL of the Azure container.
    :param blob_name: The name of the blob to create.
    :param data: The dictionary to upload.
    :return: True if upload is successful, False otherwise.
    """
    try:
        # Create a ContainerClient
        container_client = ContainerClient.from_container_url(az_container_url)

        # Create a blob client
        blob_client = container_client.get_blob_client(blob_name)

        # Convert the dictionary to a JSON string
        json_data = json.dumps(data)

        # Upload the JSON string to the blob
        blob_client.upload_blob(json_data, overwrite=True)



    except Exception as e:
        logger.exception(e)


"""
The exceptions for the crosslang module
"""

class CompilationError(Exception):
    """
    Exception raised when failed to compile the user script
    """
    def __init__(self, file, reason):
        self.file = file
        self.reason = reason

    def __str__(self):
        return "Failed to compile '{}':\n{}".format(self.file, self.reason)

class MLClientExecutionError(Exception):
    """
    Exception raised when an error occurred while running non-python ml script
    """
    def __init__(self, message):
        """
        Constructor
        """
        self.message = message

    def __str__(self):
        return self.message



from .ext_lang_map import EXTESION_LANG_MAP

import importlib
import os.path

def compile_script(script_full_path):
    """
    Compile the script to an executable according to its file extension

    This function will load the corresponding language module according to
    `EXTENSION_LANG_MAP` for compiling the script.

    @param script_full_path The full path of the target script
    @return A list of command segments for executing the executable
    """
    path_no_ext, extension = os.path.splitext(script_full_path)
    compilation_module = importlib.import_module(
        ".compile.{}.main".format(EXTESION_LANG_MAP[extension]), __package__)

    script_execution_cmd = compilation_module.compile_script(script_full_path)
    if not isinstance(script_execution_cmd, list):
        raise TypeError("The returned execution command is not a list.")

    return script_execution_cmd



from os import path

import pygame

IS_DEBUG = False
# TODO remove width and height setting
'''width and height'''
WINDOW_WIDTH = 1000
WINDOW_HEIGHT = 700

'''environment data'''
FPS = 30
SHOOT_COOLDOWN = 15

'''color'''
BLACK = "#000000"
WHITE = "#ffffff"
RED = "#ff0000"
YELLOW = "#ffff00"
GREEN = "#00ff00"
DARKGREEN = "#006400"
GREY = "#8c8c8c"
BLUE = "#0000ff"
LIGHT_BLUE = "#21A1F1"
CYAN_BLUE = "#00FFFF"
PINK = "#FF00FF"
DARKGREY = "#282828"
LIGHTGREY = "#646464"
BROWN = "#643705"
FOREST = "#22390A"
MAGENTA = "#FF00FF"
MEDGRAY = "#4B4B4B"
ORANGE =  "#FFA500"

'''command'''
TURN_LEFT_CMD = "TURN_LEFT"
TURN_RIGHT_CMD = "TURN_RIGHT"
FORWARD_CMD = "FORWARD"
BACKWARD_CMD = "BACKWARD"
AIM_LEFT_CMD = "AIM_LEFT"
AIM_RIGHT_CMD = "AIM_RIGHT"
SHOOT = "SHOOT"

'''data path'''
GAME_DIR = path.dirname(__file__)
IMAGE_DIR = path.join(GAME_DIR, "..", "asset", "image")
SOUND_DIR = path.join(GAME_DIR, "..", "asset", "sound")
MAP_DIR = path.join(GAME_DIR, "..", "asset", 'maps')

'''BG View'''
TITLE = "TankMan!"
BG_COLOR = DARKGREY
TILE_X_SIZE = 50
TILE_Y_SIZE = 50
TILE_SIZE = 50
TEXT_SIZE = 100

'''object size'''
ALL_OBJECT_SIZE = pygame.Rect(0, 0, 50, 50)
BULLET_SIZE = (13, 16)
LIVES_SIZE = (30, 25)

"""all setting"""
DOWN_IMG = 'down'
RIGHT_IMG = 'right'
UP_IMG = 'up'
LEFT_IMG = 'left'

"""collide setting"""
WITH_PLAYER = 'player'

"""map data numbers"""
PLAYER_1_IMG_NO = 1
PLAYER_2_IMG_NO = 2
WALL_IMG_NO = 3
BULLET_STATION_IMG_NO = 4
OIL_STATION_IMG_NO = 5
WALL_LIVE = 4
BULLET_SPEED = 30
BULLET_TRAVEL_DISTANCE = 300  

"""music"""
BGM = 'background_music.ogg/.wav/.mp3'
MENU_SND = 'MenuTheme.ogg/.wav/.mp3'




from os import path

import pygame
from mlgame.view.view_model import create_asset_init_data, create_image_view_data

from .env import WINDOW_HEIGHT, WINDOW_WIDTH, IMAGE_DIR, BULLET_SIZE


Vec = pygame.math.Vector2


class Bullet(pygame.sprite.Sprite):
    def __init__(self, construction, **kwargs):
        super().__init__()
        self.id = construction["_id"]
        self.no = construction["_no"]
        self.rect = pygame.Rect((0, 0), construction["_init_size"])
        self.rect.center = construction["_init_pos"]
        self.rot = kwargs["rot"]
        self.play_rect_area = kwargs["play_rect_area"]
        self.speed = kwargs["bullet_speed"] * construction["_init_size"][0] / BULLET_SIZE[0]
        self.map_width = WINDOW_WIDTH
        self.map_height = WINDOW_HEIGHT
        self.angle = 3.14 / 180 * (self.rot + 90)
        # Refactor
        if 7 > self.angle > 6:
            self.angle = 0
        self.move = {"left_up": Vec(-self.speed, -self.speed), "right_up": Vec(self.speed, -self.speed),
                     "left_down": Vec(-self.speed, self.speed), "right_down": Vec(self.speed, self.speed),
                     "left": Vec(-self.speed, 0), "right": Vec(self.speed, 0), "up": Vec(0, -self.speed),
                     "down": Vec(0, self.speed)}

                
        self.max_travel_distance = (kwargs["bullet_travel_distance"] // self.speed + 1) * self.speed

        self.travel_distance = 0

    def update(self):
        self.travel_distance += self.speed

        if self.play_rect_area.top < self.rect.centery < self.play_rect_area.bottom \
                and self.play_rect_area.left < self.rect.centerx < self.play_rect_area.right:
            is_out = False
        else:
            is_out = True

        if is_out or self.travel_distance >= self.max_travel_distance:
            self.kill()

        if self.rot == 0 or self.rot == 360:
            self.rect.center += self.move["left"]
        elif self.rot == 315 or self.rot == -45:
            self.rect.center += self.move["left_up"]
        elif self.rot == 270 or self.rot == -90:
            self.rect.center += self.move["up"]
        elif self.rot == 225 or self.rot == -135:
            self.rect.center += self.move["right_up"]
        elif self.rot == 180 or self.rot == -180:
            self.rect.center += self.move["right"]
        elif self.rot == 135 or self.rot == -225:
            self.rect.center += self.move["right_down"]
        elif self.rot == 90 or self.rot == -270:
            self.rect.center += self.move["down"]
        elif self.rot == 45 or self.rot == -315:
            self.rect.center += self.move["left_down"]

    def get_obj_progress_data(self):
        img_id = "team_a_bullet" if self.id == 1 else "team_b_bullet"
        return create_image_view_data(img_id, self.rect.x, self.rect.y, self.rect.width, self.rect.height,
                                      self.angle)

    def get_data_from_obj_to_game(self) -> dict:
        info = {"id": f"{self.no}P_bullet",
                "x": self.rect.x,
                "y": self.rect.y,
                "speed": self.speed,
                "rot": self.rot
                }
        return info



import time
from multiprocessing import Process, Pipe

from mlgame.argument.model import GroupAI
from mlgame.executor.az_uploader import AzureUploader
from mlgame.executor.recorder import ProgressLogExecutor
from mlgame.executor.websocket_executor import WebSocketExecutor
from mlgame.core.communication import GameCommManager, MLCommManager, TransitionCommManager
from mlgame.core.env import TIMEOUT
from mlgame.executor.ai_client import AIClientExecutor
from mlgame.utils.logger import logger


def create_process_of_ws_and_start(game_comm: GameCommManager, ws_url) -> Process:
    recv_pipe_for_game, send_pipe_for_ws = Pipe(False)
    recv_pipe_for_ws, send_pipe_for_game = Pipe(False)
    ws_comm = TransitionCommManager(recv_pipe_for_ws, send_pipe_for_ws,label="ws")
    game_comm.add_comm_to_others("ws", recv_pipe_for_game, send_pipe_for_game)
    ws_executor = WebSocketExecutor(ws_uri=ws_url, ws_comm=ws_comm)
    process = Process(target=ws_executor.run, name="ws")
    # process = ws_executor
    process.start()
    # time.sleep(0.1)
    return process


def create_process_of_ai_clients_and_start(
        game_comm: GameCommManager, ai_clients: list[GroupAI], game_params: dict) -> list:
    """
    return a process list to main process and bind pipes to `game_comm`
    """
    ai_process = []
    for index, ai_client in enumerate(ai_clients):
        ai_group = ai_client.group
        ai_name = ai_client.ai_name
        recv_pipe_for_game, send_pipe_for_ml = Pipe(False)
        recv_pipe_for_ml, send_pipe_for_game = Pipe(False)
        game_comm.add_comm_to_ml(
            ai_name,
            recv_pipe_for_game, send_pipe_for_game)
        ai_comm = MLCommManager(ai_name)
        ai_comm.set_comm_to_game(
            recv_pipe_for_ml, send_pipe_for_ml)
        ai_executor = AIClientExecutor(
            str(ai_client.ai_path), ai_comm,
            ai_name=ai_name,group=ai_group, game_params=game_params,ai_label=ai_client.ai_label)
        process = Process(target=ai_executor.run,
                          name=ai_name)
        process.start()
        ai_process.append(process)
    return ai_process


def create_process_of_recorder_and_start(game_comm: GameCommManager, progress_folder,
                                             progress_frame_frequency) -> Process:
    recv_pipe_for_game, send_pipe_for_pl = Pipe(False)
    recv_pipe_for_pl, send_pipe_for_game = Pipe(False)
    pl_comm = TransitionCommManager(recv_pipe_for_pl, send_pipe_for_pl,label="pl")
    game_comm.add_comm_to_others("pl", recv_pipe_for_game, send_pipe_for_game)
    pl_executor = ProgressLogExecutor(progress_folder=progress_folder,
                                      progress_frame_frequency=progress_frame_frequency, pl_comm=pl_comm)
    process = Process(target=pl_executor.run, name="pl")
    process.start()
    # time.sleep(0.1)
    return process


def create_process_of_az_uploader_and_start(game_comm: GameCommManager, az_url) -> Process:
    recv_pipe_for_game, send_pipe_for_pl = Pipe(False)
    recv_pipe_for_pl, send_pipe_for_game = Pipe(False)
    pl_comm = TransitionCommManager(recv_pipe_for_pl, send_pipe_for_pl,label="az")
    game_comm.add_comm_to_others("az", recv_pipe_for_game, send_pipe_for_game)
    az_excutor = AzureUploader(pl_comm=pl_comm, az_blob_url=az_url)
    process = Process(target=az_excutor.run, name="az")
    process.start()
    # time.sleep(0.1)
    return process

def terminate(game_comm: GameCommManager, ai_process: list, ws_proc: Process, progress_proc: Process,az_proc:Process):
    logger.debug("Main process will terminate ai process")
    # 5.terminate
    for ai_proc in ai_process:
        # Send stop signal to all alive ml processes
        if ai_proc.is_alive():
            game_comm.send_to_ml(
                None, ai_proc.name)
            ai_proc.terminate()
    logger.debug("Main process will terminate ws process")

    if ws_proc:
        timeout = time.time() + TIMEOUT
        logger.info(f"wait to close ws for timeout : {TIMEOUT} s")
        ws_proc.terminate()
        while ws_proc.is_alive():
            time.sleep(0.5)
            if time.time() > timeout:
                logger.info("Force to terminate ws proc ")
                ws_proc.kill()
                ws_proc.join()
                break

            logger.info("wait to close ws .")
        logger.info(f"use {time.time() - timeout + TIMEOUT} to close.")

    if progress_proc:
        timeout = time.time() + TIMEOUT
        logger.info(f"wait to close progress for timeout : {TIMEOUT} s")
        while True:
            time.sleep(0.5)
            progress_proc.terminate()
            if time.time() > timeout:
                print("Force to terminate progress_proc proc ")
                progress_proc.kill()
                progress_proc.join()
                break
            elif not progress_proc.is_alive():
                break
            logger.info("wait to close progress_proc .")
        logger.info(f"use {time.time() - timeout + TIMEOUT} to close.")
    if az_proc:
        timeout = time.time() + TIMEOUT
        logger.info(f"wait to close progress for timeout : {TIMEOUT} s")
        while True:
            time.sleep(0.5)
            az_proc.terminate()
            if time.time() > timeout:
                print("Force to terminate az_proc ")
                az_proc.kill()
                az_proc.join()
                break
            elif not az_proc.is_alive():
                break
            logger.info("wait to close az_proc .")
        logger.info(f"use {time.time() - timeout + TIMEOUT} to close.")
    logger.debug("Game is terminated")



import pygame


class MLPlay:
    def __init__(self,*args,**kwargs):
        self.other_cars_position = []
        self.coins_pos = []
        print("Initial ml script")

    def update(self, scene_info: dict, keyboard=[], *args, **kwargs):
        """
        Generate the command according to the received scene information
        """
        # print(scene_info)
        if scene_info["status"] != "GAME_ALIVE":
            return "RESET"

        command = []
        if pygame.K_LEFT in keyboard:
            command.append("BRAKE")
        if pygame.K_RIGHT in keyboard:
            command.append("SPEED")
        if pygame.K_UP in keyboard:
            command.append("MOVE_LEFT")
        if pygame.K_DOWN in keyboard:
            command.append("MOVE_RIGHT")

        return command
    def reset(self):
        """
        Reset the status
        """
        # print("reset ml script")
        pass



import json

import pytest
from loguru import logger
from pydantic import ValidationError

from mlgame.core.exceptions import ErrorEnum
from mlgame.core.model import (
    MLGameEntityWrapperSchema,
    GameProgressSchema,
    GameInfoSchema,
    GameErrorSchema,
    MLGameDataType, SystemMsgSchema,
)
from mlgame.view.view_model import get_dummy_progress_data


def test_game_progress_schema():
    data = [
        {
            "frame": 1,
            "background": [],
            "object_list": [
                {
                    "type": "image",
                    "x": 490,
                    "y": 391,
                    "width": 10,
                    "height": 10,
                    "image_id": "ball",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 400,
                    "width": 40,
                    "height": 10,
                    "image_id": "board",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 400,
                    "y": 50,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 425,
                    "y": 50,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 50,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 500,
                    "y": 50,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 550,
                    "y": 50,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 575,
                    "y": 50,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 400,
                    "y": 60,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 575,
                    "y": 60,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 575,
                    "y": 70,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 425,
                    "y": 80,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 450,
                    "y": 80,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 500,
                    "y": 80,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 525,
                    "y": 80,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 575,
                    "y": 80,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 575,
                    "y": 90,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 575,
                    "y": 100,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 400,
                    "y": 110,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 110,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 550,
                    "y": 110,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 400,
                    "y": 120,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 550,
                    "y": 120,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 425,
                    "y": 130,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 450,
                    "y": 130,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 130,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 500,
                    "y": 130,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 525,
                    "y": 130,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 550,
                    "y": 130,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 140,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 400,
                    "y": 150,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 150,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 550,
                    "y": 150,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 575,
                    "y": 150,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 425,
                    "y": 160,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 160,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 525,
                    "y": 160,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 450,
                    "y": 170,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 170,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 500,
                    "y": 170,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 180,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 190,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 450,
                    "y": 200,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 500,
                    "y": 200,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 425,
                    "y": 210,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 525,
                    "y": 210,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 400,
                    "y": 220,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 550,
                    "y": 220,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 575,
                    "y": 220,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
            ],
            "toggle_with_bias": [],
            "toggle": [],
            "foreground": [
                {
                    "type": "text",
                    "content": "catching ball: 0",
                    "color": "#FFFFFF",
                    "x": 430,
                    "y": 479,
                    "font-style": "16px Arial",
                },
                {
                    "type": "text",
                    "content": "remain brick: 36",
                    "color": "#FFFFFF",
                    "x": 430,
                    "y": 459,
                    "font-style": "16px Arial",
                },
                {
                    "type": "text",
                    "content": "remain hard brick: 11",
                    "color": "#FFFFFF",
                    "x": 430,
                    "y": 439,
                    "font-style": "16px Arial",
                },
            ],
            "user_info": [],
            "game_sys_info": {},
        },
        {
            "frame": 2,
            "background": [],
            "object_list": [
                {
                    "type": "image",
                    "x": 483,
                    "y": 384,
                    "width": 10,
                    "height": 10,
                    "image_id": "ball",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 470,
                    "y": 400,
                    "width": 40,
                    "height": 10,
                    "image_id": "board",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 400,
                    "y": 50,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 425,
                    "y": 50,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 50,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 500,
                    "y": 50,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 550,
                    "y": 50,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 575,
                    "y": 50,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 400,
                    "y": 60,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 575,
                    "y": 60,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 575,
                    "y": 70,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 425,
                    "y": 80,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 450,
                    "y": 80,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 500,
                    "y": 80,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 525,
                    "y": 80,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 575,
                    "y": 80,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 575,
                    "y": 90,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 575,
                    "y": 100,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 400,
                    "y": 110,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 110,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 550,
                    "y": 110,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 400,
                    "y": 120,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 550,
                    "y": 120,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 425,
                    "y": 130,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 450,
                    "y": 130,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 130,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 500,
                    "y": 130,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 525,
                    "y": 130,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 550,
                    "y": 130,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 140,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 400,
                    "y": 150,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 150,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 550,
                    "y": 150,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 575,
                    "y": 150,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 425,
                    "y": 160,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 160,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 525,
                    "y": 160,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 450,
                    "y": 170,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 170,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 500,
                    "y": 170,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 180,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 475,
                    "y": 190,
                    "width": 25,
                    "height": 10,
                    "image_id": "hard_brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 450,
                    "y": 200,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 500,
                    "y": 200,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 425,
                    "y": 210,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 525,
                    "y": 210,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 400,
                    "y": 220,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 550,
                    "y": 220,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
                {
                    "type": "image",
                    "x": 575,
                    "y": 220,
                    "width": 25,
                    "height": 10,
                    "image_id": "brick",
                    "angle": 0,
                },
            ],
            "toggle_with_bias": [],
            "toggle": [],
            "foreground": [
                {
                    "type": "text",
                    "content": "catching ball: 0",
                    "color": "#FFFFFF",
                    "x": 430,
                    "y": 479,
                    "font-style": "16px Arial",
                },
                {
                    "type": "text",
                    "content": "remain brick: 36",
                    "color": "#FFFFFF",
                    "x": 430,
                    "y": 459,
                    "font-style": "16px Arial",
                },
                {
                    "type": "text",
                    "content": "remain hard brick: 11",
                    "color": "#FFFFFF",
                    "x": 430,
                    "y": 439,
                    "font-style": "16px Arial",
                },
            ],
            "user_info": [],
            "game_sys_info": {},
        },
    ]
    progress = []
    for obj in data:
        progress.append(GameProgressSchema.model_validate(obj).__dict__)
    dummy_progress =get_dummy_progress_data()
    key_of_dummy = dummy_progress.keys()
    assert key_of_dummy == progress[0].keys()

    invalid_data = {
        "frame": "invalid_frame",
        "background": "invalid_background",
        "object_list": "invalid_object_list",
        "toggle": "invalid_toggle",
        "toggle_with_bias": "invalid_toggle_with_bias",
        "foreground": "invalid_foreground",
        "game_sys_info": "invalid_game_sys_info",
        "musics": "invalid_musics",
        "sounds": "invalid_sounds"
    }

    with pytest.raises(ValidationError):
        GameProgressSchema(**invalid_data)


def test_game_info_schema():
    game_info ={"scene": {"color": "#54AFE3", "width": 1000, "bias_x": 0, "bias_y": 0, "height": 500}, "assets": [{"url": "https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/arkanoid/main/asset/img/brick.png", "type": "image", "width": 25, "height": 10, "image_id": "brick", "file_path": "/game/src/../asset/img/brick.png"}, {"url": "https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/arkanoid/main/asset/img/hard_brick.png", "type": "image", "width": 25, "height": 10, "image_id": "hard_brick", "file_path": "/game/src/../asset/img/hard_brick.png"}, {"url": "https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/arkanoid/main/asset/img/ball.png", "type": "image", "width": 5, "height": 5, "image_id": "ball", "file_path": "/game/src/../asset/img/ball.png"}, {"url": "https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/arkanoid/main/asset/img/board.png", "type": "image", "width": 40, "height": 5, "image_id": "board", "file_path": "/game/src/../asset/img/board.png"}, {"url": "https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/arkanoid/main/asset/img/bg.png", "type": "image", "width": 1000, "height": 500, "image_id": "bg", "file_path": "/game/src/../asset/img/bg.png"}], "background": [{"x": 0, "y": 0, "type": "image", "angle": 0, "width": 1000, "height": 500, "image_id": "bg"}],"musics":[],"sounds":[] }
    validated_obj = GameInfoSchema.model_validate(game_info).__dict__

    invalid_game_info = {
        "scene": "invalid_scene",
        "assets": "invalid_assets",
        "background": "invalid_background",
        "musics": "invalid_musics",
        "sounds": "invalid_sounds"
    }

    with pytest.raises(ValidationError):
        GameInfoSchema(**invalid_game_info)
    # assert game_info.keys() == validated_obj.keys()
    pass

def test_ml_game_entity_wrapper_schema():
    # Test with GameProgressSchema
    progress_data = GameProgressSchema(frame=1)
    wrapper = MLGameEntityWrapperSchema(data=progress_data)
    assert wrapper.type == MLGameDataType.GAME_PROGRESS

    # Test with GameInfoSchema
    info_data = GameInfoSchema(scene={"key": "value"})
    wrapper = MLGameEntityWrapperSchema(data=info_data)
    assert wrapper.type == MLGameDataType.GAME_INFO

    # Test with GameErrorSchema
    error_data = GameErrorSchema(message="An error occurred", error_type=ErrorEnum.GAME_EXEC_ERROR, frame=123)
    wrapper = MLGameEntityWrapperSchema(data=error_data)
    assert wrapper.type == MLGameDataType.GAME_ERROR

    # Test with invalid data type
    with pytest.raises(ValueError):
        wrapper = MLGameEntityWrapperSchema(data="Invalid data type")

def test_json_dumps_entity():
    obj = MLGameEntityWrapperSchema(
        type=MLGameDataType.SYSTEM_MSG,
        data=SystemMsgSchema(message="system")
    )
    logger.info(obj)
    json_str = obj.model_dump_json()
    logger.info(json_str)
    assert json.loads(json_str)

    obj = MLGameEntityWrapperSchema(
        type=MLGameDataType.GAME_INFO,
        # data=GameInfoSchema(scene={})
        data=GameErrorSchema(error_type=ErrorEnum.GAME_EXEC_ERROR, frame=1)
    )
    assert obj.type == MLGameDataType.GAME_ERROR
    logger.info(obj)
    json_str = obj.model_dump_json()
    logger.info(json_str)


    obj = MLGameEntityWrapperSchema(
        type=MLGameDataType.GAME_INFO,
        data=GameInfoSchema(scene={})
    )
    assert obj.type == MLGameDataType.GAME_INFO
    logger.info(obj)
    json_str = obj.model_dump_json()
    logger.info(json_str)
    result = obj.model_dump(mode='json')
    logger.info(result)

def test_game_info():
    obj = GameInfoSchema(scene={"key": "value"})
    logger.info(obj)
    json_str = obj.model_dump_json()

    logger.info(json_str)


from datetime import datetime

from mlgame.utils.azure import upload_data_to_azure_blob, upload_data_to_azure_blob_by_container
from .env import AZURE_CONTAINER_URL, AZURE_BLOB_URL


def test_upload_to_azure_blob():
    current_time = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
    blob_name = f"upload/blob_{current_time}.json"
    data = {"key": "value"}
    upload_data_to_azure_blob_by_container(AZURE_CONTAINER_URL, blob_name, [data])


def test_create_new_file_in_azure_blob():
    file_name = "03.json"
    data = {"key": "value"}
    
    upload_data_to_azure_blob(AZURE_BLOB_URL, file_name, [data])


import random
from .env import *


class Scene():
    def __init__(self, width: int, height: int, color: str = "#000000"):
        """
        This is a value object
        :param width:
        :param height:
        :param color:
        :param image:
        """
        self.width = width
        self.height = height
        self.color = color


def get_progress_data(game_mode):
    game_progress = {
        "background": [],
        "object_list": [],
        "toggle": [],
        "foreground": [],
        "user_info": [],
        "game_sys_info": {}
    }

    for car in game_mode.car_info:
        if car["id"] % 2 == 0:
            x = 600
        else:
            x = 730

        if car["status"]:
            game_progress["background"].append(
                get_dummy_text("L:" + str(car["l_sensor_value"]["distance"]) + "cm", "#FFFF00", (x,
                                                                                                 178 + 20 + 94 * (car[
                                                                                                                      "id"] // 2)),
                               "15px Arial"))
            game_progress["background"].append(
                get_dummy_text("F:" + str(car["f_sensor_value"]["distance"]) + "cm", "#FF0000", (x,
                                                                                                 178 + 40 + 94 * (car[
                                                                                                                      "id"] // 2)),
                               "15px Arial"))
            game_progress["background"].append(
                get_dummy_text("R:" + str(car["r_sensor_value"]["distance"]) + "cm", "#21A1F1", (x,
                                                                                                 178 + 60 + 94 * (car[
                                                                                                                      "id"] // 2)),
                               "15px Arial"))
            game_progress["object_list"].append(get_line_object("l_sensor", car["center"],
                                                                trnsfer_box2d_to_pygame(game_mode,
                                                                                        car["l_sensor_value"][
                                                                                            "coordinate"]), "#FFFF00",
                                                                5))

            game_progress["object_list"].append(get_line_object("l_top_sensor", car["center"],
                                                                trnsfer_box2d_to_pygame(game_mode,
                                                                                        car["l_t_sensor_value"][
                                                                                            "coordinate"]), "#FFFF00",
                                                                5))

            game_progress["object_list"].append(get_line_object("r_top_sensor", car["center"],
                                                                trnsfer_box2d_to_pygame(game_mode,
                                                                                        car["r_t_sensor_value"][
                                                                                            "coordinate"]), "#21A1F1",
                                                                5))
            game_progress["object_list"].append(get_line_object("r_sensor", car["center"],
                                                                trnsfer_box2d_to_pygame(game_mode,
                                                                                        car["r_sensor_value"][
                                                                                            "coordinate"]), "#21A1F1",
                                                                5))
            game_progress["object_list"].append(get_line_object("f_sensor", car["center"],
                                                                trnsfer_box2d_to_pygame(game_mode,
                                                                                        car["f_sensor_value"][
                                                                                            "coordinate"]), "#FF0000",
                                                                5))
        else:
            game_progress["background"].append(get_dummy_text(str(car["end_frame"]) + "frame", "#FFFFFF",
                                                              (x, 178 + 40 + 94 * (car["id"] // 2)), "15px Arial"))
    # try:
    #     for car in game_mode.car_info:
    #         game_progress["object_list"].append(
    #             get_image_object("car_0"+str(car["id"]+1), (car["topleft"][0], car["topleft"][1]), 50, 40, car["angle"])
    #         )
    # except Exception:
    #     pass
    for user in game_mode.car_info:
        game_progress["user_info"].append({
            "player_num": str(user["id"] + 1) + "P",
            "F_sensor": user["f_sensor_value"]["distance"],
            "R_sensor": user["r_sensor_value"]["distance"],
            "L_sensor": user["l_sensor_value"]["distance"],
            "R_PWM": user["R_PWM"],
            "L_PWM": user["L_PWM"],
        })

    game_progress["game_sys_info"]["frame"] = game_mode.frame
    return game_progress


def get_scene_init_sample_data() -> dict:
    """
    :rtype: dict
    :return:  遊戲場景初始化的資料
    """

    scene = Scene(WIDTH, HEIGHT)
    assets = [
        {
            "type": "image",
            "image_id": 'car_01',
            "width": 50,
            "height": 50,
            "url": 'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/main/game_core/image/car_01.png'
        }, {
            "type": "image",
            "image_id": 'car_02',
            "width": 50,
            "height": 50,
            "url": 'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/main/game_core/image/car_02.png'
        }, {
            "type": "image",
            "image_id": 'car_03',
            "width": 50,
            "height": 50,
            "url": 'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/main/game_core/image/car_03.png'
        }, {
            "type": "image",
            "image_id": 'car_04',
            "width": 50,
            "height": 50,
            "url": 'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/main/game_core/image/car_04.png'
        }, {
            "type": "image",
            "image_id": 'car_05',
            "width": 50,
            "height": 50,
            "url": 'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/main/game_core/image/car_05.png'
        }, {
            "type": "image",
            "image_id": 'car_06',
            "width": 50,
            "height": 50,
            "url": 'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/main/game_core/image/car_06.png'
        }, {
            "type": "image",
            "image_id": 'info',
            "width": 306,
            "height": 480,
            "url": 'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/main/game_core/image/info.png'
        }, {
            "type": "image",
            "image_id": 'target',
            "width": 40,
            "height": 40,
            "url": 'https://raw.githubusercontent.com/PAIA-Playful-AI-Arena/Maze_Car/main/game_core/image/target.png'
        }
    ]
    return {"scene": scene.__dict__,
            "assets": assets,
            # "audios": {}
            }


def get_image_object(image_id, coordinate, width, height, angle=0):
    """
    這是一個用來繪製圖片的資料格式，
    "type"表示不同的類型
    "x" "y" 表示物體左上角的座標
    "width" "height"表示其大小
    "image_id"表示其圖片的識別號，需在
    "angle"表示其順時針旋轉的角度
    """
    return {"type": "image",
            "x": coordinate[0],
            "y": coordinate[1],
            "width": width,
            "height": height,
            "image_id": image_id,
            "angle": int(angle)}


def get_rect_object(name, coordinate, width, height, color, angle=0):
    """
    這是一個用來繪製矩形的資料格式，
    "type"表示不同的類型
    "x""y"表示其位置，位置表示物體左上角的座標
    "size"表示其大小
    "image"表示其圖片
    "angle"表示其順時針旋轉的角度
    "color"以字串表示
    :return:
    """
    return {"type": "rect",
            "name": name,
            "x": coordinate[0],
            "y": coordinate[1],
            "angle": int(angle),
            "width": width,
            "height": height,
            "color": color
            }


def get_line_object(name, dot1, dot2, color, width=2):
    """
    這是一個用來繪製矩形的資料格式，
    "type"表示不同的類型
    "x""y"表示其位置，位置表示物體左上角的座標
    "size"表示其大小
    "image"表示其圖片
    "angle"表示其順時針旋轉的角度
    "color"以字串表示
    :return:
    """
    return {"type": "line",
            "name": name,
            "x1": int(dot1[0]),
            "y1": int(dot1[1]),
            "x2": int(dot2[0]),
            "y2": int(dot2[1]),
            "width": width,
            "color": color
            }


def get_polygon_object(name, points, color):
    """
    這是一個用來繪製多邊形的資料格式，
    points欄位至少三個 # [{"x":1,"y":2},{},{}]
    :return:dict
    """
    vertices = []
    for p in points:
        vertices.append({"x": p[0], "y": p[1]})
    return {"type": "polygon",
            "name": name,
            "color": color,
            "points": vertices
            }


def get_dummy_text(content, color, coordinate, font_style="24px Arial"):
    return {
        "type": "text",
        "content": content,
        "color": color,
        "x": coordinate[0],
        "y": coordinate[1],
        "font-style": font_style
    }


def get_dummy_progress_data():
    pass
    # return {"game_object_list":
    #             [get_dummy_polygon(), get_dummy_image(),
    #              get_dummy_rect(), get_dummy_text()],
    #         "game_user_info":
    #             [{"player": "1P",
    #               "key": "value"
    #               }, {}, {}],
    #         "game_sys_info": {}
    #         }


def get_dummy_result_data():
    return {
        "frame_used": 15,
        "ranks": [
            [  # 1st
                {"player_num": "1P",  # 給系統mapping
                 "game_result": "7s",  # 顯示給玩家的
                 # 可自行增加
                 },
                {"player_num": "1P",  # 給系統mapping
                 "game_result": "7s",  # 顯示給玩家的
                 # 可自行增加
                 }
            ],
            [  # 2nd

            ],
            [  # 3rd

            ]
            #     player
        ],

    }


def gen_points(point_num: int = 4) -> list:
    result = []
    for i in range(point_num):
        result.append({"x": random.randint(0, 100), "y": random.randint(0, 100)})
    return result


def gen_rects(rect_num: int = 1) -> list:
    result = []
    for i in range(rect_num):
        result.append(gen_points(4))
    return result


def trnsfer_box2d_to_pygame(game, coordinate):
    '''
    :param coordinate: vertice of body of box2d object
    :return: center of pygame rect
    '''
    return ((coordinate[0] - game.pygame_point[0]) * PPM, (game.pygame_point[1] - coordinate[1]) * PPM)



import logging

from mlgame.utils.logger import get_singleton_logger


def test_get_single_logger():
    logger = get_singleton_logger()
    assert isinstance(logger,logging.Logger)

    # logger.debug("debug fadsfasdf")


import datetime
import os
from enum import Enum
from pathlib import Path
from typing import Union, List, Optional

from pydantic import AnyUrl, BaseModel, FilePath, DirectoryPath, field_validator, model_validator

from mlgame.core.env import FPS_MAX
from mlgame.utils.io import check_folder_existed_and_readable_or_create
from mlgame.utils.logger import logger


class GroupEnum(str, Enum):
    A = "A"
    B = "B"
    C = "C"
    D = "D"
    O = "O"


class AINameEnum(str, Enum):
    P1 = "1P"
    P2 = "2P"
    P3 = "3P"
    P4 = "4P"
    P5 = "5P"
    P6 = "6P"
    P7 = "7P"
    P8 = "8P"

class GroupAI(BaseModel):
    group:GroupEnum
    ai_name:AINameEnum
    ai_path: FilePath
    ai_label:str = ""
    
    @model_validator(mode="after")
    def set_ai_label_if_empty(self):
        if not self.ai_label:
            self.ai_label = self.ai_name
        return self


class MLGameArgument(BaseModel):
    """
    Data Entity to handle parsed cli arguments
    """
    fps: int = 30
    progress_frame_frequency: int = 300
    one_shot_mode: bool = False
    # ai_clients: Optional[List[FilePath]] = None
    is_manual: bool = False
    no_display: bool = True
    ws_url: Optional[AnyUrl] = None
    az_upload_url: Optional[AnyUrl] = None
    game_folder: DirectoryPath
    game_params: List[str]
    output_folder: Union[Path, None] = None
    record_folder: Union[Path, None] = None
    is_sound_on : bool = True
    is_debug: bool = False
    group_ai_orig: Optional[List[str]] = None
    labeled_ai_orig: Optional[List[str]] = None
    

    @property
    def group_ai(self)-> List[GroupAI]:
        group_ai = []
        if self.group_ai_orig:
            for entry in self.group_ai_orig:
                group, ai_name, ai_path,ai_label = entry.split(",")
                group_ai.append(GroupAI(group=group, ai_name=ai_name, ai_path=Path(ai_path), ai_label=ai_label))
        elif self.labeled_ai_orig:
            for i,ai_client_str in enumerate(self.labeled_ai_orig,start=1):
                if "," in ai_client_str:
                    ai_path,ai_label = ai_client_str.split(",")
                else:
                    ai_path = Path(ai_client_str)
                    ai_label = AINameEnum(f"{i}P")
                # group, ai_name, ai_path = entry.split(",")
                group_ai.append(GroupAI(group=GroupEnum.O, ai_name=AINameEnum(f"{i}P"), ai_path=Path(ai_path),ai_label=ai_label))
        return group_ai


    @field_validator('is_manual', mode='before')
    def update_manual(cls, v, values) -> bool:
        return not ("ai_clients" in values or "group_ai_orig" in values)

    @field_validator('output_folder', 'record_folder', mode='before')
    def update_folder(cls, v, values):
        if v is None:
            return None
        path = os.path.join(
            str(v),
            datetime.datetime.now().strftime('%Y%m%d-%H%M%S')
        )
        if check_folder_existed_and_readable_or_create(path):
            return path

    @field_validator('fps', mode='before')
    def clamp_fps(cls, v):
        fps_max = FPS_MAX
        fps_min = 1
        try:
            v = int(v)
            if v< fps_min or v> fps_max:
                logger.warning(
                    f"FPS should be between {fps_min} and {fps_max}, system will automatically clamp the fps"
                )
        except Exception:
            return 30  # fallback to default if not int
        return max(fps_min, min(fps_max, v))


class UserNumConfig(BaseModel):
    """
    Data Entity to handle user_num in game_config.json
    """
    min: int
    max: int

    @model_validator(mode="after")
    def max_should_be_larger_than_min(cls,values):
        assert values.min <= values.max
        return values



import time

from mlgame.argument.model import AINameEnum, GroupEnum
from mlgame.core.communication import MLCommManager
from mlgame.core.exceptions import ErrorEnum, GameError
from mlgame.executor.interface import ExecutorInterface
from mlgame.utils.logger import logger


import importlib
import os
import traceback


class AIClientExecutor(ExecutorInterface):
    def __init__(self,
                 # TODO insert GroupAI
                 ai_client_path: str,
                 ai_comm: MLCommManager,
                 ai_name:AINameEnum=AINameEnum.P1,
                 group:GroupEnum=GroupEnum.A,
                 game_params:dict=None,
                 ai_label:str=""
                 ):
        if game_params is None:
            game_params = {}
        self._frame_count = 0
        self.ai_comm = ai_comm
        self.ai_path = ai_client_path
        self._proc_name = ai_client_path
        # self._args_for_ml_play = args
        # self._kwargs_for_ml_play = kwargs
        self._ai_name = ai_name
        self._group = group
        self._ai_label = ai_label
        self.game_params = game_params

    def run(self):
        # self.ai_comm.start_recv_obj_thread()
        try:
            module_name = os.path.basename(self.ai_path)
            spec = importlib.util.spec_from_file_location(
                module_name, self.ai_path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            # set ai_name and group name
            ai_obj = module.MLPlay(ai_name=self._ai_name.value,
                                   # TODO ai_label
                                   group=self._group.value,
                                   game_params=self.game_params,
                                   ai_label=self._ai_label)

            # cmd = ai_obj.update({})
            logger.info(f"             AI Client({self._ai_name}) runs")
            self._ml_ready()
            while True:
                data = self.ai_comm.recv_from_game() # non-blocking
                # TODO add schema in ai client executor
                if data:
                    scene_info, keyboard_info,ai_enabled = data
                    if scene_info is None:
                        # game over
                        break
                else:
                    print(f"ai receive from game : {data}")
                    break

                # assert keyboard_info == "1"
                if ai_enabled:
                    # 遊戲暫停、delay 等等，都不會進到此條件，也因此確保了拿到的遊戲資訊是最新的
                    command = ai_obj.update(scene_info, keyboard_info)
                    if scene_info["status"] != "GAME_ALIVE" or command == "RESET":
                        ai_obj.reset()
                        # TODO 收到 reset 這裡會重置 frame_count ，應該要再多確認什麼情況要重置，什麼情況不用重置
                        self._frame_count = 0
                        self._ml_ready()
                        continue
                    if command is not None:
                        action_frame = scene_info['frame']
                        logger.debug(f"frame count in ai : {action_frame}")
                        self.ai_comm.send_to_game({
                            "frame": action_frame,
                            "command": command
                        })

                # time.sleep(0.01)
        except ModuleNotFoundError as e:
            failed_module_name = e.__str__().split("'")[1]
            logger.error(
                f"Module '{failed_module_name}' is not found in {self._proc_name}")
                        # send msg to game process
            ai_error = GameError(
                error_type=ErrorEnum.AI_INIT_ERROR, frame=self._frame_count,
                message="The process '{}' AI is not imported correctly. {}".format(
                    self._ai_name, traceback.format_exc())
            )

            self.ai_comm.send_to_game(ai_error)
            self._wait_for_exit()

        except FileNotFoundError as e:

            logger.error(f"{e} in {self._proc_name}")
            # send msg to game process
            ai_error = GameError(
                error_type=ErrorEnum.AI_INIT_ERROR, frame=self._frame_count,
                message=f"The process '{self._ai_name}' AI occurred file not found error. {e.__str__()}"
            )

            self.ai_comm.send_to_game(ai_error)
            self._wait_for_exit()

        except Exception as e:
            # handle ai other error

            logger.error(f"{e} in {self._proc_name}")
            ai_error = GameError(
                error_type=ErrorEnum.AI_EXEC_ERROR, frame=self._frame_count,
                message=f"The process '{self._ai_name}' is exited by itself. {traceback.format_exc()}"
            )

            self.ai_comm.send_to_game(ai_error)
            self._wait_for_exit()

        except SystemExit:  # Catch the exception made by 'sys.exit()'
            logger.error("             System exit at ai client process ")

            ai_error = GameError(
                error_type=ErrorEnum.AI_EXEC_ERROR, frame=self._frame_count,
                message=f"The process '{self._ai_name}' is exited by sys.exit. : {traceback.format_exc()}"
            )

            self.ai_comm.send_to_game(ai_error)

        if module == "mlgame.crosslang.ml_play":
            # TODO crosslang
            ai_obj.stop_client()
        logger.info(f"             AI Client({self._ai_name}) ends")

    def _ml_ready(self):
        """
        Send a "READY" command to the game process
        """
        self.ai_comm.send_to_game("READY")

    def _wait_for_exit(self):
        logger.warning(f"             AI Client({self._ai_name}) is broken , will return None until the end")

        while True:
            # mock AI client, will not return anything
            data = self.ai_comm.recv_from_game() # non-blocking
            if data:
                scene_info, keyboard_info,ai_enabled = data
                if scene_info is None:
                    # game over
                    break
            else:
                print(f"ai receive from game : {data}")
                break

            # assert keyboard_info == "1"
            if ai_enabled:
                # 遊戲暫停、delay 等等，都不會進到此條件，也因此確保了拿到的遊戲資訊是最新的
                if scene_info["status"] != "GAME_ALIVE" :
                    # ai_obj.reset()
                    self._ml_ready()
                    continue
                command = None
                action_frame = scene_info['frame']
                logger.debug(f"frame count in ai : {action_frame}")
                self.ai_comm.send_to_game({
                    "frame": action_frame,
                    "command": command
                })
            time.sleep(1)



from datetime import datetime, timezone
from enum import Enum
from typing import Any, Dict, List, Union

from pydantic import BaseModel, model_validator

from mlgame.core.exceptions import ErrorEnum


class MLGameDataType(str, Enum):
    """Data type enum for MLGame"""

    GAME_PROGRESS = "game_progress"
    GAME_ERROR = "game_error"
    GAME_RESULT = "game_result"
    SYSTEM_MSG = "system_message"
    DEFAULT = "not_yet_set"
    GAME_INFO = "game_info"
    NONE = "none"
    END = "end"

class GameProgressSchema(BaseModel):
    """Data structure for game progress"""

    frame: int
    background: List[Dict[str, Any]] = []
    object_list: List[Dict[str, Any]] = []
    toggle: List[Dict[str, Any]] = []
    toggle_with_bias: List[Dict[str, Any]] = []
    foreground: List[Dict[str, Any]] = []
    game_sys_info: Dict[str, Any] = {}
    musics: List[Dict[str, Any]] = []
    sounds: List[Dict[str, Any]] = []



class GameInfoSchema(BaseModel):
    """Data structure for game initialization"""

    scene: Dict[str, Any]
    assets: List[Dict[str, Any]] = []
    background: List[Dict[str, Any]] = []
    musics: List[Dict[str, Any]] = []
    sounds: List[Dict[str, Any]] = []

class GameResultSchema(BaseModel):
    """Data structure for game result"""
    frame_used:int
    status:str
    attachment:List[Dict[str, Any]]

class GameErrorSchema(BaseModel):
    """Data structure for game errors"""

    error_type: ErrorEnum
    message: str = ""
    frame: int
    time_stamp: datetime = datetime.now(timezone.utc)


    class Config:
        arbitrary_types_allowed = True

class SystemMsgSchema(BaseModel):
    """Data structure for system messages"""

    message: str = ""


    class Config:
        arbitrary_types_allowed = True
    def __dict__(self):
        return {"message": self.message}


class MLGameEntityWrapperSchema(BaseModel):
    """Base schema for MLGame data exchange"""

    type: MLGameDataType = MLGameDataType.DEFAULT
    data: Union[GameProgressSchema, GameInfoSchema,GameResultSchema, GameErrorSchema,SystemMsgSchema,Dict[str, Any]]={}

    @model_validator(mode="after")
    def update_type_from_data(self):
        """Update the `type` field after the model is fully validated."""
        if isinstance(self.data, GameProgressSchema):
            self.type = MLGameDataType.GAME_PROGRESS
        elif isinstance(self.data, GameInfoSchema):
            self.type = MLGameDataType.GAME_INFO
        elif isinstance(self.data, GameResultSchema):
            self.type = MLGameDataType.GAME_RESULT
        elif isinstance(self.data, GameErrorSchema):
            self.type = MLGameDataType.GAME_ERROR
        elif isinstance(self.data, SystemMsgSchema):
            self.type = MLGameDataType.SYSTEM_MSG
        

        return self

    # class Config:
    #     arbitrary_types_allowed = True



import logging
import os.path

from os import path
from mlgame.argument.tool import get_data_from_json_file
from mlgame.argument.game_argument import parse_game_config_data, GameConfig


def test_create_game_config():
    game_folder = os.path.join(os.path.dirname(__file__), "..", "..", "games", "swimming_squid_battle")
    game_config = GameConfig(game_folder=game_folder)
    # game_config._config_to_create_parser
    assert hasattr(game_config, 'game_config_parser')
    assert hasattr(game_config, 'game_version')
    assert hasattr(game_config, 'game_cls')

    assert game_config.logo
    assert os.path.exists(game_config.logo)
    

    game_params = "--level 3 --game_times 3".split(' ')
    parsed_game_params = game_config.parse_game_params(game_params)
    assert isinstance(parsed_game_params, dict)


def test_parse_game_config_file():
    config_file = path.join(path.dirname(__file__), "test_data", "game_config.json")
    config_data = get_data_from_json_file(config_file)
    params = config_data["game_params"]

    result = parse_game_config_data(config_data)
    assert result
    assert "()" in result
    for param in params:
        assert param["name"] in result
    logging.info(result)



import json
import re
from functools import wraps

import pydantic

from mlgame.game.paia_game import GameResultState
from mlgame.view.audio_model import MusicProgressSchema, SoundProgressSchema, MusicInitSchema, SoundInitSchema

K_ATTACHMENT = "attachment"
K_STATUS = "status"
K_ASSET = "assets"
K_BACKGROUND="background"

OBJ_TYPE_TXT = "text"
OBJ_TYPE_IMG = "image"
OBJ_TYPE_RECT = "rect"
OBJ_TYPE_POLY = "polygon"
OBJ_TYPE_LINE = "line"

K_FRAME_USED = "frame_used"
K_URL = "url"
K_IMG_ID = "image_id"
K_IMG = "images"
K_TEMPLATE = "template"
K_Game_Objs = "object_list"
K_SCENE = "scene"
K_SIZE = "size"
K_ANGLE = "angle"
K_COLOR = "color"
K_CORD = "coordinate"
K_NAME = 'name'
K_TYPE = 'type'
K_X = 'x'
K_Y = 'y'
K_WID = 'width'
K_HEIGHT = 'height'
K_POINT = 'points'
K_POLY = 'polys'
K_STYLE = 'style'


def assert_polygon_obj(polygon):
    assert_contains_keys(polygon, keys=[K_TYPE, K_NAME, K_POINT, K_COLOR])
    assert type(polygon[K_POINT]) == list
    assert len(polygon[K_POINT]) >= 3
    assert type(polygon[K_POINT][0]) == dict
    assert_contains_keys(polygon[K_POINT][0], [K_X, K_Y])
    # validate the color string
    assert_color_rgba_hex_string(polygon[K_COLOR])


def assert_rect_obj(rect):
    assert isinstance(rect[K_X], (int, float))
    assert isinstance(rect[K_Y], (int, float))
    assert isinstance(rect[K_WID], (int, float))
    assert isinstance(rect[K_HEIGHT], (int, float))

    # validate the color string
    assert_color_rgba_hex_string(rect[K_COLOR])


def assert_image_obj(img):
    assert_contains_keys(img, keys=[K_TYPE, K_X, K_Y, K_WID, K_HEIGHT, K_ANGLE, K_IMG_ID])
    assert isinstance(img[K_X], (int, float))
    assert isinstance(img[K_Y], (int, float))
    assert isinstance(img[K_WID], (int, float))
    assert isinstance(img[K_HEIGHT], (int, float))
    assert isinstance(img[K_ANGLE], (int, float))


def assert_line_obj(line):
    assert_contains_keys(line, keys=[K_TYPE, "x1", "y1", "x2", "y2", K_WID, K_COLOR])
    assert isinstance(line["x1"], (int, float))
    assert isinstance(line["y1"], (int, float))
    assert isinstance(line["x2"], (int, float))
    assert isinstance(line["y2"], (int, float))
    assert isinstance(line[K_WID], (int, float))
    assert_color_rgba_hex_string(line[K_COLOR])


def assert_scene_init_data(data: dict = None):
    """  驗證遊戲場景初始化的資料
    :return:None
    """
    assert type(data) == dict ,data
    assert_contains_keys(data, [K_SCENE, K_ASSET,K_BACKGROUND])

    scene_obj = data[K_SCENE]
    assert type(scene_obj) == dict
    assert_contains_keys(scene_obj, [K_WID, K_HEIGHT, K_COLOR])

    images = data[K_ASSET]
    assert type(images) == list
    assert_contains_keys(images[0], [K_IMG_ID, K_URL, K_WID, K_HEIGHT])
    # TODO check background

    musics = data.get("musics",[])
    for music in musics:
        assert_init_music_obj(music)

    sounds = data.get("sounds",[])
    for sound in sounds:
        assert_init_sound_obj(sound)
def assert_contains_keys(obj: dict, keys: list):
    for k in keys:
        assert k in obj, str(obj.__class__) + " should contains " + str(keys)
    pass


def assert_color_rgba_hex_string(color_str: str = ""):
    assert re.match(r'^#(\d|[A-F]|[a-f]){6,8}$', color_str), "color should be in hex format ex #DDEEFF00"


def assert_text_obj(text_obj: dict):
    assert_contains_keys(text_obj, keys=["content", K_COLOR, K_X, K_Y, "font-style"])
    assert isinstance(text_obj[K_X], (int, float))
    assert isinstance(text_obj[K_Y], (int, float))
    assert type(text_obj["font-style"]) == str
    assert_color_rgba_hex_string(text_obj[K_COLOR])


def assert_music_obj(music):
    assert 'type' in music
    MusicProgressSchema(**music)
    pass
def assert_init_music_obj(music):
    assert 'type' in music
    MusicInitSchema(**music)
    pass


def assert_sound_obj(sound):
    assert 'type' in sound
    SoundProgressSchema(**sound)
    pass

def assert_init_sound_obj(sound):
    assert 'type' in sound
    SoundInitSchema(**sound)
    pass


def assert_game_progress_data(data: dict = None):
    """
    驗證遊戲過程的更新資料
    :param data:
    :return:
    """

    assert type(data) == dict
    assert_contains_keys(data,
                         ["frame", "background", "object_list", "toggle", "foreground", "user_info", "game_sys_info"])
    assert type(data[K_Game_Objs]) == list
    total_objs = []
    total_objs.extend(data[K_Game_Objs])
    total_objs.extend(data["background"])
    total_objs.extend(data["toggle"])
    total_objs.extend(data["foreground"])
    for obj in total_objs:
        if obj[K_TYPE] == OBJ_TYPE_POLY:
            assert_polygon_obj(obj)
        elif obj[K_TYPE] == OBJ_TYPE_IMG:
            assert_image_obj(obj)
        elif obj[K_TYPE] == OBJ_TYPE_RECT:
            assert_rect_obj(obj)
        elif obj[K_TYPE] == OBJ_TYPE_TXT:
            assert_text_obj(obj)
        elif obj[K_TYPE] == OBJ_TYPE_LINE:
            assert_line_obj(obj)
        pass

    musics = data.get("musics",[])
    for music in musics:
        assert_music_obj(music)

    sounds = data.get("sounds",[])
    for sound in sounds:
        assert_sound_obj(sound)

    # TODO if contains 'music' or 'sound' , should be in specific format

def assert_game_result_data(data: dict = None):
    """
        驗證遊戲過程的更新資料
        :param data:
        :return:
        """
    # data = get_dummy_progress_data()
    assert type(data) == dict
    assert_contains_keys(data, [K_FRAME_USED, K_STATUS, K_ATTACHMENT])
    assert data[K_STATUS] in GameResultState.__dict__.values()

    assert isinstance(data[K_ATTACHMENT], list)
    for obj in data[K_ATTACHMENT]:
        assert_contains_keys(obj,['rank','player_num'])


def check_game_result(func):
    """
    這是用來檢驗 遊戲結果 的裝飾子，可以在開發期間使用，在正式執行環境再拿掉
    :param func:
    :return:
    """
    # TODO use env to decide whether assert or not
    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        assert_game_result_data(result)
        return result

    return wrapper


def check_scene_init_data(func):
    """
    這是用來檢驗 遊戲初始化資料 的裝飾子，可以在開發期間使用，在正式執行環境再拿掉
    :param func:
    :return:
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        assert_scene_init_data(result)
        return result

    return wrapper


def check_game_progress(func):
    """
    這是用來檢驗 遊戲過程資料 的裝飾子，可以在開發期間使用，在正式執行環境再拿掉
    :param func:
    :return:
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        assert_game_progress_data(result)
        return result

    return wrapper



from __future__ import annotations

import sys

from mlgame.core.security import install_exec_killer
sys.path.append('.')

import time
from typing import Optional, Sequence

from mlgame.argument.cmd_argument import parse_cmd_and_get_arg_obj
from mlgame.argument.tool import revise_ai_clients
from mlgame.core.process import create_process_of_az_uploader_and_start
from mlgame.executor.game import GameExecutor
from mlgame.executor.manual import GameManualExecutor
from mlgame.game.paia_game import get_paia_game_obj
from mlgame.utils.logger import logger
from mlgame.view.audio_model import MusicInitSchema, SoundInitSchema
from mlgame.view.sound_controller import SoundController
from mlgame.argument.model import MLGameArgument
from mlgame.argument.game_argument import GameConfig

from mlgame.core.communication import GameCommManager
from mlgame.core.process import (
    create_process_of_ai_clients_and_start,
    create_process_of_ws_and_start,
    create_process_of_recorder_and_start,
    terminate,
)
from mlgame.view.view import PygameView, DummyPygameView


class MLGameObj:
    def __init__(self, args: MLGameArgument):
        self.args = args

        self._ai_processes: list = []
        self._ws_proc = None
        self._record_proc = None
        self._az_upload_proc = None
        self._game_executor: Optional[GameExecutor | GameManualExecutor] = None
        self._game_comm: Optional[GameCommManager] = None


        game_config = GameConfig(game_folder=self.args.game_folder.__str__())
        parsed_game_params = game_config.parse_game_params(game_params=self.args.game_params)

        ai_clients = revise_ai_clients(
            ai_clients=self.args.group_ai,
            user_num_config=game_config.user_num_config
        )
        user_num = len(ai_clients)
        game = get_paia_game_obj(
            game_cls=game_config.game_cls,
            parsed_game_params=parsed_game_params,
            user_num=user_num,
            group_ai_list=ai_clients
        )

        self._game_config = game_config
        self._parsed_game_params = parsed_game_params
        self._ai_clients = ai_clients
        self._game = game

        if getattr(self.args, "is_debug", False):
            logger.remove()
            logger.add(
                sys.stdout,
                level="DEBUG",
                format="<green>{time:YYYY-MM-DD HH:mm:ss.SSSSS}</green> | "
                       "<level>{level: <8}</level> | <cyan>{name}</cyan>:"
                       "<cyan>{function}</cyan>:<cyan>{line}</cyan> - "
                       "<level>{message}</level>"
            )
            logger.add(
                "debug.log",
                level="DEBUG",
                rotation="10 MB",
                retention=1,
                compression=None,
            )

        self._setup_runtime()

    def _setup_runtime(self):
        logger.info("===========Game is started ===========")
        self._game_comm = GameCommManager()

        if self.args.ws_url:
            self._ws_proc = create_process_of_ws_and_start(self._game_comm, str(self.args.ws_url))

        if self.args.record_folder:
            self._record_proc = create_process_of_recorder_and_start(
                self._game_comm, self.args.record_folder, self.args.progress_frame_frequency
            )

        if self.args.az_upload_url:
            self._az_upload_proc = create_process_of_az_uploader_and_start(
                self._game_comm, self.args.az_upload_url
            )

        init_data = self._game.get_scene_init_data()



        if self.args.is_manual:
            game_view = PygameView(
                init_data,
                caption=f"PAIA Game: {self._game_config.game_name} v{self._game_config.game_version}",
                icon=self._game_config.logo,
            )
            self._game_executor = GameManualExecutor(
                self._game,
                game_view,
                self._game_comm,
                fps=getattr(self.args, "fps", 30),
                one_shot_mode=getattr(self.args, "one_shot_mode", False),
            )
        else:
            sound_controller = SoundController(
                is_sound_on=self.args.is_sound_on,
                music_objs=[MusicInitSchema(**obj) for obj in init_data.get("musics", [])],
                sound_objs=[SoundInitSchema(**obj) for obj in init_data.get("sounds", [])],
            )
            if self.args.no_display:
                logger.warning("Game will not be displayed.")
                game_view = DummyPygameView(init_data)
            else:
                game_view = PygameView(
                    init_data,
                    caption=f"PAIA Game: {self._game_config.game_name} v{self._game_config.game_version}",
                    icon=self._game_config.logo,
                    sound_controller=sound_controller,
                )

            self._ai_processes = create_process_of_ai_clients_and_start(
                game_comm=self._game_comm,
                ai_clients=self._ai_clients,
                game_params=self._parsed_game_params,
            )
            self._game_executor = GameExecutor(
                self._game,
                self._game_comm,
                game_view,
                fps=getattr(self.args, "fps", 30),
                one_shot_mode=getattr(self.args, "one_shot_mode", False),
                no_display=self.args.no_display,
                output_folder=getattr(self.args, "output_folder", None),
            )

    def run(self):
        install_exec_killer()

        if not self._game_executor:
            raise RuntimeError("Game executor not initialized.")
        time.sleep(0.1)
        try:
            self._game_executor.run()
        except Exception as e:
            logger.exception(e)
            raise
        finally:
            self.close()

    def close(self):
        if self._game_comm is None:
            return
        terminate(self._game_comm, self._ai_processes, self._ws_proc, self._record_proc, self._az_upload_proc)
        self._game_comm = None
        logger.info("===========All process is terminated ===========")


def run_mlgame(arg_obj: MLGameArgument):
    logger.info("===========Game is started ===========")
    install_exec_killer()
    game = MLGameObj(arg_obj)
    game.run()

def main(argv: Optional[Sequence[str]] = None):
    argv = list(sys.argv[1:] if argv is None else argv)
    arg_obj = parse_cmd_and_get_arg_obj(argv)
    run_mlgame(arg_obj)

if __name__ == "__main__":
    import os
    os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "hide"

    main()


"""
The template of the main script of the machine learning process
"""
import random
import pygame

from src.env import IS_DEBUG


class MLPlay:
    def __init__(self, ai_name, *args, **kwargs):
        """
        Constructor

        @param side A string "1P" or "2P" indicates that the `MLPlay` is used by
               which side.
        """
        self.side = ai_name
        print(f"Initial Game {ai_name} ml script")
        self.time = 0

    def update(self, scene_info: dict, keyboard=[], *args, **kwargs):
        """
        Generate the command according to the received scene information
        """
        # print(keyboard)
        
        if scene_info["status"] != "GAME_ALIVE":
            # print(scene_info)
            return "RESET"
        move_act = random.randrange(5)
        aim_act = random.randrange(3)
        shoot_cd = random.randrange(15, 31)

        is_shoot = 0
        if scene_info["used_frame"] % shoot_cd == 0:
            is_shoot = random.randrange(2)

        command = []
        if move_act == 1:
            command.append("TURN_RIGHT")
        elif move_act == 2:
            command.append("TURN_LEFT")
        elif move_act == 3:
            command.append("FORWARD")
        elif move_act == 4:
            command.append("BACKWARD")

        if aim_act == 1:
            command.append("AIM_LEFT")
        elif aim_act == 2:
            command.append("AIM_RIGHT")

        if is_shoot and not IS_DEBUG:
            command.append("SHOOT")

        if self.side == "1P":
            if pygame.K_b in keyboard:
                command.append("DEBUG")

        if not command:
            command.append("NONE")
        
        return command

    def reset(self):
        """
        Reset the status
        """
        print(f"reset Game {self.side}")



"""
This is a base class for different mode in game.
"""
from .car import Car
from .points import *
from .maze_wall import SlantWall

import pygame

from .env import *


class GameMode(object):
    def __init__(self, bg_img=pygame.Surface((WIDTH, HEIGHT))):
        self.bg_img = bg_img
        self.clock = pygame.time.Clock()
        self.running = True
        self.frame = 0
        pygame.font.init()
        self.font = pygame.font.Font(pygame.font.match_font("arial", bold=True), 15)
        self.time_font = pygame.font.Font(pygame.font.match_font("arial", bold=True), 46)
        self.check_point_num = 0
        self.check_points = []
        # self.start_time = time.time()

    def ticks(self, fps=FPS):
        """This method should be called once per frame.
        It will compute how many milliseconds have passed since the previous call.
        :param fps: frame per second 每秒的繪圖次數
        :return: None
        """
        self.clock.tick(fps)

    def handle_event(self):
        """ Handle the event from window , mouse or button.
        :return: None
        """
        pass

    def detect_collision(self):
        """ Detect the collision event between sprites.
        :return: None
        """
        pass

    def update_sprite(self, *args):
        """ This function should update every sprite in games.
        :return: None
        """
        pass

    def draw_bg(self):
        """  Draw a background on screen.
        :return:None
        """
        pass

    def drawWorld(self):
        """  This function should draw every sprite on specific surface.
        :return: None
        """
        pass

    def flip(self):
        """Update the full display Surface to the screen
        :return:None
        """
        pygame.display.flip()

    def isRunning(self) -> bool:
        return self.running

    def rank(self):
        '''
        不限碰撞次數，但須在規定時間內過關
        排名依據：
        1. 檢查點個數（走到終點及取得所有檢查點）。
        2. 若檢查點個數相同，則比較碰撞次數，越少者排名越前。
        3. 碰撞次數相同，則比較走到最末檢查點的時間，越早走到者排名越前。
        '''
        user_check_point = []
        user_crush_times = []
        user_end_frame = []
        user_score = []
        for car in self.eliminated_user:
            user_check_point.append(car.check_point)
            user_crush_times.append(car.collide_times)
            user_end_frame.append(car.end_frame)
            user_score.append(car.check_point * 10000 - car.collide_times * 10 - car.end_frame * 0.001)
        rank_user = []  # [sprite]
        while user_score:
            max_score = max(user_score)
            max_index = user_score.index(max_score)
            rank_user.append(self.eliminated_user[max_index])
            user_score.pop(max_index)
            self.eliminated_user.pop(max_index)
        # result = [user_score.index(x) for x in sorted(user_score, reverse=True)]
        # for i in range(len(result)):
        #     rank_user.append(self.eliminated_user[result[i]])
        return rank_user

    def pygame_to_box2d(self, coodinate, map_h):
        '''
        只用在終點跟檢查點，因為他們不屬於 Box2d 的物件，但需要提供資訊
        '''
        return (coodinate[0]/PPM*5,(map_h -  coodinate[1] / PPM)*5)

    def transfer_box2d_to_pygame(self, coordinate):
        '''
        :param coordinate: vertice of body of box2d object
        :return: center of pygame rect
        '''
        return ((coordinate[0] - self.pygame_point[0]) * PPM +160, (self.pygame_point[1] - coordinate[1]) * PPM-960)

    def get_wall_info_v(self, wall_tile):
        wall_tiles = []
        for col in range(len(self.map.data[0]) - 1):
            row = 0
            first_tile = -1
            last_tile = -1
            while row < len(self.map.data):
                tiles = self.map.data[row]

                if (tiles[col] % 18) == wall_tile:
                    if first_tile == -1:
                        first_tile = row
                        if row == len(self.map.data) - 1:
                            last_tile = row
                            self.wall_vertices_for_Box2D.append(
                                {"type": wall_tile,
                                 "vertices": self.wall_vertices_v((col, first_tile), (col, last_tile))})
                            first_tile = -1
                            row += 1
                        else:
                            row += 1
                    elif row == len(self.map.data) - 1:
                        last_tile = row
                        self.wall_vertices_for_Box2D.append(
                            {"type": wall_tile, "vertices": self.wall_vertices_v((col, first_tile), (col, last_tile))})
                        first_tile = -1
                        row += 1
                    else:
                        row += 1
                else:
                    if first_tile != -1:
                        last_tile = row - 1
                        self.wall_vertices_for_Box2D.append(
                            {"type": wall_tile, "vertices": self.wall_vertices_v((col, first_tile), (col, last_tile))})
                        # self.wall_vertices_for_Box2D.append(self.wall_vertices_v((col, first_tile), (col, last_tile)))
                        first_tile = -1
                        row += 1
                    else:
                        row += 1

    def get_wall_info_h(self, wall_tile):
        wall_tiles = []
        for row, tiles in enumerate(self.map.data):
            col = 0
            first_tile = -1
            last_tile = -1
            while col < (len(tiles)):
                if (tiles[col] % 18) == wall_tile:
                    if first_tile == -1:
                        first_tile = col
                        if col == len(tiles) - 1:
                            first_tile = -1
                            col += 1
                        else:
                            col += 1
                    elif col == len(tiles) - 1:
                        last_tile = col
                        self.wall_vertices_for_Box2D.append(
                            {"type": wall_tile, "vertices": self.wall_vertices_h((first_tile, row), (last_tile, row))})
                        # self.wall_vertices_for_Box2D.append(self.wall_vertices_h((first_tile, row), (last_tile, row)))
                        for i in range(first_tile, last_tile + 1):
                            tiles[i] = 0
                        first_tile = -1
                        col += 1
                    else:
                        col += 1
                else:
                    if first_tile != -1:
                        last_tile = col - 1
                        if first_tile == last_tile:
                            first_tile = -1
                            col += 1
                        else:
                            self.wall_vertices_for_Box2D.append(
                                {"type": wall_tile,
                                 "vertices": self.wall_vertices_h((first_tile, row), (last_tile, row))})
                            for i in range(first_tile, last_tile + 1):
                                tiles[i] = 0
                            first_tile = -1
                            col += 1
                    else:
                        col += 1

    def _print_result(self):
        if self.is_end and self.x == 0:
            for user in self.ranked_user:
                self.result.append(str(user.car_no + 1) + "P:" + str(user.end_frame) + "frame")
            self.x += 1

    def load_map_object(self, obj):
        o = obj["end_point"]
        # tiled -> box2d 需要對調 x y
        x, y = self.transfer_box2d_to_pygame((o[1], o[0]))
        self.end_point = End_point(self, (x, y))
        self.check_point_num += 1
        # o = obj["check_point"]
        # print(o)
            # self.check_point_num += 1
        o = obj["car"]
        #
        if o[2] == 3:
            for world in self.worlds:
                car = Car(world, (o[1], o[0]), self.worlds.index(world), self.sensor_num, 2)
                self.cars.add(car)
                self.car_info.append(car.get_info())
        elif o[2] == 4:
            for world in self.worlds:
                car = Car(world, (o[1], o[0]), self.worlds.index(world), self.sensor_num, 0.5)
                self.cars.add(car)
                self.car_info.append(car.get_info())
        elif o[2] == 2:
            for world in self.worlds:
                car = Car(world, (o[1], o[0]), self.worlds.index(world), self.sensor_num, 1)
                self.cars.add(car)
                self.car_info.append(car.get_info())
        elif o[2] == 5:
            for world in self.worlds:
                car = Car(world, (o[1], o[0]), self.worlds.index(world), self.sensor_num, 1.5)
                self.cars.add(car)
                self.car_info.append(car.get_info())
        try:
            if self.end_point and len(self.cars):
                pass
            else:
                print("Map without car")
                self.running = False
        except:
            print("Map without end point")
            self.running = False



from mlgame.core.communication import GameCommManager
from mlgame.executor.interface import ExecutorInterface
from mlgame.game.generic import quit_or_esc
from mlgame.game.paia_game import PaiaGame
from mlgame.utils.logger import logger
from mlgame.view.view import PygameViewInterface


import pandas as pd


import time
import traceback


class GameManualExecutor(ExecutorInterface):
    # TODO deprecated
    def __init__(self, game: PaiaGame,
                 game_view: PygameViewInterface,
                 game_comm: GameCommManager,
                 fps=30,
                 one_shot_mode=False, ):
        self.game_view = game_view
        self.frame_count = 0
        self.game = game
        self.game_comm = game_comm
        self._ml_delayed_frames = {}
        self._ml_execution_time = 1 / fps
        self._fps = fps
        self._ml_delayed_frames = {}
        # self._recorder = get_recorder(self._execution_cmd, self._ml_names)
        self._frame_count = 0
        self.one_shot_mode = one_shot_mode
        self._proc_name = self.game.__class__.__str__

    def run(self):
        game = self.game
        game_view = self.game_view
        self.game_comm.send_to_others(game_view.scene_init_data)

        try:
            while not quit_or_esc():
                cmd_dict = game.get_keyboard_command()
                # self._recorder.record(scene_info_dict, cmd_dict)
                result = game.update(cmd_dict)
                self._frame_count += 1
                view_data = game.get_scene_progress_data()
                self.game_comm.send_to_others(view_data)
                game_view.draw(view_data)
                time.sleep(self._ml_execution_time)
                # Do reset stuff
                if result == "RESET" or result == "QUIT":
                    game_result = game.get_game_result()
                    attachments = game_result['attachment']
                    print(pd.DataFrame(attachments).to_string())
                    if self.one_shot_mode or result == "QUIT":
                        game_result['frame_used']=self._frame_count
                        self.game_comm.send_to_others(game_result)
                        break
                    game.reset()
                    game_view.reset()
                    # self._frame_count = 0

        except Exception as e:
            # handle unknown exception
            # send to es
            logger.exception(
                f"Some errors happened in game process. {e.__str__()}")
            traceback.print_exc()

        logger.info("manual executor end.")


import abc
import os.path
import time
from functools import lru_cache

import math
import pygame

from mlgame.core.model import GameProgressSchema
from mlgame.view.decorator import K_BACKGROUND, K_SCENE
from mlgame.view.sound_controller import SoundController

KEYS = [
    pygame.K_a, pygame.K_b, pygame.K_c, pygame.K_d, pygame.K_e, pygame.K_f, pygame.K_g, pygame.K_h, pygame.K_i,
    pygame.K_j, pygame.K_k, pygame.K_l, pygame.K_m, pygame.K_n, pygame.K_o, pygame.K_p, pygame.K_q, pygame.K_r,
    pygame.K_s, pygame.K_t, pygame.K_u, pygame.K_v, pygame.K_w, pygame.K_x, pygame.K_y, pygame.K_z,
    pygame.K_1, pygame.K_2, pygame.K_3, pygame.K_4, pygame.K_5,
    pygame.K_6, pygame.K_7, pygame.K_8, pygame.K_9, pygame.K_0,
    pygame.K_UP, pygame.K_DOWN, pygame.K_LEFT, pygame.K_RIGHT,
]

LINE = "line"
TEXT = "text"
NAME = "name"
TYPE = "type"
ANGLE = "angle"
SIZE = "size"
COLOR = "color"
IMAGE = "image"
RECTANGLE = "rect"
POLYGON = "polygon"
AAPOLYGON = "aapolygon"


@lru_cache
def transfer_hex_to_rgb(hex_str: str) -> tuple:
    h = hex_str.lstrip('#')
    return tuple(int(h[i:i + 2], 16) for i in (0, 2, 4))


@lru_cache
def transfer_hex_to_rgba(hex_str: str) -> tuple:
    temp_str = str(hex_str)
    if len(hex_str) == 7:
        temp_str += "FF"
    h = temp_str.lstrip('#')
    return tuple(int(h[i:i + 2], 16) for i in (0, 2, 4, 6))


@lru_cache
def scale_bias_of_coordinate(obj_length, scale):
    return obj_length / 2 * (1 - scale)


@lru_cache
def rotate_img(scaled_img, radian_angle):
    return pygame.transform.rotate(
        scaled_img,
        (radian_angle * 180 / math.pi) % 360
    )


@lru_cache
def scale_img(img, origin_width, origin_height, scale_ratio):
    return pygame.transform.scale(
        img, (int(origin_width * scale_ratio), int(origin_height * scale_ratio))
    )


class PygameViewInterface(abc.ABC):
    def __init__(self, game_info: dict,sound_controller:SoundController):
        self.sound_controller = sound_controller
        pass

    @abc.abstractmethod
    def reset(self):
        pass

    @abc.abstractmethod
    def draw(self, game_progress: GameProgressSchema):
        pass

    @abc.abstractmethod
    def get_keyboard_info(self) -> list:
        return []

    @abc.abstractmethod
    def save_image(self, img_path: os.path.abspath):
        pass
    @abc.abstractmethod
    def play_audio(self,game_progress: GameProgressSchema ):
        pass

    def is_paused(self):
        return False


class DummyPygameView(PygameViewInterface):
    def __init__(self, game_info: dict, sound_controller:SoundController=None):
        super().__init__(game_info,None)

    def reset(self):
        pass

    def draw(self, game_progress: GameProgressSchema):
        pass

    def save_image(self, img_path: os.path.abspath):
        pass

    def get_keyboard_info(self) -> list:
        return []
    def play_audio(self,game_progress: GameProgressSchema):
        pass

def create_rounded_icon(image_path, size=(80, 80), radius=20):
    """Loads an image, applies a rounded border, and returns the modified surface."""
    icon = pygame.image.load(image_path)  # Load image
    icon = pygame.transform.smoothscale(icon, size)  # Resize to ensure it's the right size

    # Create a mask with transparency
    mask = pygame.Surface(size, pygame.SRCALPHA)  # Alpha for transparency
    pygame.draw.rect(mask, (255, 255, 255, 255), (0, 0, *size), border_radius=radius)  # White rounded rect

    # Apply mask to make the corners transparent
    icon.blit(mask, (0, 0), special_flags=pygame.BLEND_RGBA_MULT)  # Merge the mask

    return icon

class PygameView(PygameViewInterface):
    EMBEDED_FONT ={
        "burnfont":os.path.join(os.path.dirname(__file__),"..","assets", "burnfont.otf"),
        "NotoSansTC":os.path.join(os.path.dirname(__file__),"..","assets", "NotoSansTC-Regular.ttf")
    }
    def __init__(self, game_info: dict, caption="PAIA Game",icon=None,sound_controller:SoundController=None, *args, **kwargs):
        super().__init__(game_info,sound_controller)
        self._pause_state = False
        self._last_pause_btn_clicked_time = 0
        pygame.display.init()
        self._caption = caption
        pygame.display.set_caption(caption)
        if icon:
            icon_img = create_rounded_icon(icon)
        else:
            icon_img = create_rounded_icon(os.path.join(os.path.dirname(__file__),"..","assets", "icon.png"))
        pygame.display.set_icon(icon_img)

        pygame.font.init()
        self.scene_init_data = game_info
        self.width = self.scene_init_data[K_SCENE]["width"]
        self.height = self.scene_init_data[K_SCENE]["height"]
        self.background_color = transfer_hex_to_rgb(self.scene_init_data[K_SCENE][COLOR])
        self.screen = pygame.display.set_mode(
            (self.width, self.height),
            flags=pygame.RESIZABLE | pygame.SCALED)
        self.address = "GameView"
        self.image_dict = self.loading_image()
        # TODO load music and sound
        self._fixed_backgound_objs = self.scene_init_data.get(K_BACKGROUND, [])
        self.font = {}

        # self.map_width = game_info["map_width"]
        # self.map_height = game_info["map_height"]
        self.origin_bias_point = [self.scene_init_data[K_SCENE]["bias_x"], self.scene_init_data[K_SCENE]["bias_y"]]
        self.bias_point_var = [0, 0]
        self.bias_point = self.origin_bias_point.copy()

        self.scale = 1
        # if "images" in game_info.keys():
        #     self.image_dict = self.loading_image(game_info["images"])
        self._toggle_on = True
        self._toggle_last_time = 0
        self._last_frame_time = time.time()
        self._last_m_key_pressed= time.time()


    def reset(self):
        self.bias_point_var = [0, 0]
        self.bias_point = self.origin_bias_point.copy()

        self.scale = 1
        self._toggle_on = True
        self._toggle_last_time = 0

    def loading_image(self):
        result = {}
        if "assets" in self.scene_init_data:
            for file in self.scene_init_data["assets"]:
                # print(file)
                if file[TYPE] == IMAGE:
                    image = pygame.image.load(file["file_path"]).convert_alpha()
                    result[file["image_id"]] = image
        return result

    def draw(self, game_progress: GameProgressSchema):
        '''
        每個frame呼叫一次，把角色畫在螢幕上
        :param game_progress:
        :return:
        '''
        current_time = time.time()
        # prevent division by zero
        elapsed_time = current_time - self._last_frame_time +1e-6
        
        fps = 1 / elapsed_time 
        pygame.display.set_caption(f"{self._caption} FPS: {fps:.2f}")
        self._last_frame_time = current_time
        if not isinstance(game_progress, GameProgressSchema):
            game_progress =GameProgressSchema.model_validate(game_progress)
        

        self.screen.fill(self.background_color)
        self.adjust_pygame_screen()
        # TODO draw fixed frame in scene_init data

        if "view_center_coordinate" in game_progress.game_sys_info:
            self.origin_bias_point = [
                game_progress.game_sys_info["view_center_coordinate"][0],
                game_progress.game_sys_info["view_center_coordinate"][1]
            ]
            self.bias_point[0] = self.origin_bias_point[0] + self.bias_point_var[0]
            self.bias_point[1] = self.origin_bias_point[1] + self.bias_point_var[1]
        for game_object in self._fixed_backgound_objs:
            self.draw_game_obj_according_type_with_bias(game_object, self.bias_point[0], self.bias_point[1], self.scale)

        for game_object in game_progress.background:
            self.draw_game_obj_according_type_with_bias(game_object, self.bias_point[0], self.bias_point[1], self.scale)
        for game_object in game_progress.object_list:
            # let object could be shifted
            self.draw_game_obj_according_type_with_bias(game_object, self.bias_point[0], self.bias_point[1], self.scale)
        if self._toggle_on:
            for game_object in game_progress.toggle:
                self.draw_game_obj_according_type(game_object)
            for game_object in game_progress.toggle_with_bias:
                # let object could be shifted
                self.draw_game_obj_according_type_with_bias(
                    game_object, self.bias_point[0], self.bias_point[1], self.scale
                )

        for game_object in game_progress.foreground:
            # object should not be shifted
            self.draw_game_obj_according_type(game_object)
        pygame.display.flip()

    def save_image(self, img_path: os.path.abspath):
        # should do this work in separate thread
        pygame.image.save(self.screen, img_path)
        pass

    def draw_game_obj_according_type(self, game_object, scale=1):
        if game_object[TYPE] == IMAGE:
            self.draw_image(
                game_object["image_id"], game_object["x"], game_object["y"],
                game_object["width"], game_object["height"], game_object["angle"], scale)

        elif game_object[TYPE] == RECTANGLE:
            self.draw_rect(
                game_object["x"], game_object["y"], game_object["width"], game_object["height"],
                transfer_hex_to_rgba(game_object[COLOR]), scale)

        elif game_object[TYPE] == POLYGON:
            self.draw_polygon(
                game_object["points"], transfer_hex_to_rgba(game_object[COLOR]), scale)

        elif game_object[TYPE] == TEXT:
            self.draw_text(
                game_object["content"], game_object["font-style"],
                game_object["x"], game_object["y"], transfer_hex_to_rgba(game_object[COLOR]), scale)
        elif game_object[TYPE] == LINE:
            self.draw_line(
                game_object["x1"], game_object["y1"], game_object["x2"], game_object["y2"],
                game_object["width"], transfer_hex_to_rgba(game_object[COLOR]), scale)
        else:
            pass

    def draw_game_obj_according_type_with_bias(self, game_object, bias_x, bias_y, scale=1):
        if game_object[TYPE] == IMAGE:
            self.draw_image(
                game_object["image_id"], game_object["x"] + bias_x, game_object["y"] + bias_y,
                game_object["width"], game_object["height"], game_object["angle"], scale)

        elif game_object[TYPE] == RECTANGLE:
            self.draw_rect(
                game_object["x"] + bias_x, game_object["y"] + bias_y, game_object["width"],
                game_object["height"],
                transfer_hex_to_rgba(game_object[COLOR]), scale)

        elif game_object[TYPE] == POLYGON:
            self.draw_polygon(game_object["points"], transfer_hex_to_rgba(game_object[COLOR]), bias_x, bias_y, scale)
        elif game_object[TYPE] == AAPOLYGON:
            self.draw_aapolygon(game_object["points"], transfer_hex_to_rgba(game_object[COLOR]), bias_x, bias_y, scale)

        elif game_object[TYPE] == TEXT:
            self.draw_text(
                game_object["content"], game_object["font-style"],
                game_object["x"] + bias_x, game_object["y"] + bias_y,
                transfer_hex_to_rgba(game_object[COLOR]),
                scale)
        elif game_object[TYPE] == LINE:
            self.draw_line(
                game_object["x1"] + bias_x, game_object["y1"] + bias_y,
                game_object["x2"] + bias_x, game_object["y2"] + bias_y,
                game_object["width"],
                transfer_hex_to_rgba(game_object[COLOR]), scale)

        else:
            pass

        # hex # need turn to RGB

    def draw_image(self, image_id, x, y, width, height, radian_angle, scale=1):
        scaled_img = scale_img(self.image_dict[image_id], width, height, scale)
        rotated_img = rotate_img(scaled_img, radian_angle)
        # print(angle)
        rect = rotated_img.get_rect()
        rect.x = x * scale + scale_bias_of_coordinate(self.width, scale)
        rect.y = y * scale + scale_bias_of_coordinate(self.height, scale)
        self.screen.blit(rotated_img, rect)

    def draw_rect(self, x: int, y: int, width: int, height: int, color, scale=1):
        # pygame.draw.rect(
        #     self.screen, color,
        #     pygame.Rect(x * scale + scale_bias_of_coordinate(self.width, scale),
        #                 y * scale + scale_bias_of_coordinate(self.height, scale),
        #                 width * scale,
        #                 height * scale))
        transparent_surface = pygame.Surface((width * scale, height * scale), pygame.SRCALPHA)
        transparent_surface.fill(color)
        
        self.screen.blit(transparent_surface,  (x * scale + scale_bias_of_coordinate(self.width, scale),
                                 y * scale + scale_bias_of_coordinate(self.height, scale)))

    def draw_line(self, x1, y1, x2, y2, width, color, scale=1):
        # TODO revise sharper
        offset_width = scale_bias_of_coordinate(self.width, scale)
        offset_height = scale_bias_of_coordinate(self.height, scale)

        # import pygame.gfxdraw
        # pygame.gfxdraw.line(
        #     self.screen,
        #     int(x1 * scale + offset_width),int( y1 * scale + offset_height),
        #     int(x2 * scale + offset_width), int(y2 * scale + offset_height),
        #     color
        #     # int(width * scale)
        # )
        # pygame.draw.line(
        #     self.screen, color,
        #     (x1 * scale + offset_width, y1 * scale + offset_height),
        #     (x2 * scale + offset_width, y2 * scale + offset_height),
        #     1
        # )
        if scale != 1:

            offset_width = scale_bias_of_coordinate(self.width, scale)
            offset_height = scale_bias_of_coordinate(self.height, scale)
            pygame.draw.line(self.screen, color, (x1 * scale + offset_width, y1 * scale + offset_height),
                             (x2 * scale + offset_width, y2 * scale + offset_height), int(width * scale))
        else:
            pygame.draw.line(self.screen, color, (x1, y1), (x2 * scale, y2), int(width))

    def draw_polygon(self, points, color, bias_x=0, bias_y=0, scale=1):
        vertices = []
        for p in points:
            vertices.append((
                (p["x"] + bias_x) * scale + scale_bias_of_coordinate(self.width, scale),
                (p["y"] + bias_y) * scale + scale_bias_of_coordinate(self.height, scale)
            ))
        pygame.draw.polygon(self.screen, color, vertices)

    def draw_aapolygon(self, points, color, bias_x=0, bias_y=0, scale=1):
        vertices = []
        for p in points:
            vertices.append((
                (p["x"] + bias_x) * scale + scale_bias_of_coordinate(self.width, scale),
                (p["y"] + bias_y) * scale + scale_bias_of_coordinate(self.height, scale)
            ))
        # TODO use aalines
        # pygame.draw.aalines(self.screen, color=color, points=vertices,closed=True,
        #                     blend=13)
        pygame.draw.polygon(self.screen, color, vertices, width=scale * 5)
        # import pygame.gfxdraw
        # pygame.gfxdraw.aapolygon(
        #     self.screen,
        #     vertices,
        #     color
        # )
        # pygame.gfxdraw.filled_polygon(
        #     self.screen,
        #     vertices,
        #     color
        # )

    def draw_text(self, text, font_style, x, y, color, scale=1):
        if font_style in self.font.keys():
            font = self.font[font_style]
        else:
            # TODO update parser could parse size bold font-family automatically
            font_style_list = font_style.split(" ", -1)
            size = int(font_style_list[0].replace("px", "", 1))
            font_type = font_style_list[1].lower()
            if font_type in self.EMBEDED_FONT.keys():
                font = pygame.font.Font(self.EMBEDED_FONT[font_type], int(size * scale))
            else:
                font_path = pygame.font.match_font(font_type)
                font = pygame.font.Font(font_path, int(size * scale))
            if "BOLD" in font_style_list:
                font.bold=True

            self.font[font_style] = font
        text_surface = font.render(text, True, color)
        text_rect = text_surface.get_rect()
        text_rect.x, text_rect.y = (x * scale + scale_bias_of_coordinate(self.width, scale),
                                    y * scale + scale_bias_of_coordinate(self.height, scale))
        self.screen.blit(text_surface, text_rect)

    def adjust_pygame_screen(self):
        """
        zoom in zoom out and shift the window.
        """
        key_state = pygame.key.get_pressed()
        # 上下左右 放大縮小
        if key_state[pygame.K_i]:
            self.bias_point_var[1] += 10
            self.bias_point[1] = self.origin_bias_point[1] + self.bias_point_var[1]
        elif key_state[pygame.K_k]:
            self.bias_point_var[1] -= 10
            self.bias_point[1] = self.origin_bias_point[1] + self.bias_point_var[1]
        elif key_state[pygame.K_j]:
            self.bias_point_var[0] += 10
            self.bias_point[0] = self.origin_bias_point[0] + self.bias_point_var[0]
        elif key_state[pygame.K_l]:
            self.bias_point_var[0] -= 10
            self.bias_point[0] = self.origin_bias_point[0] + self.bias_point_var[0]

        if key_state[pygame.K_o]:
            self.scale += 0.01
        elif key_state[pygame.K_u]:
            self.scale -= 0.01
            if self.scale < 0.05:
                self.scale = 0.05
        # 隱藏鍵
        if key_state[pygame.K_h] and (time.time() - self._toggle_last_time) > 0.3:
            self._toggle_on = not self._toggle_on
            self._toggle_last_time = time.time()

    def get_keyboard_info(self) -> list:
        keyboard_info = []
        pressed_keys = pygame.key.get_pressed()
        if True in pressed_keys:
            for k in KEYS:
                if pressed_keys[k]:
                    keyboard_info.append(k)
        return keyboard_info

    def is_paused(self) -> bool:
        # 隱藏鍵
        key_state = pygame.key.get_pressed()
        if key_state[pygame.K_p] and (time.time() - self._last_pause_btn_clicked_time) > 0.3:
            self._pause_state = not self._pause_state
            self._last_pause_btn_clicked_time = time.time()
        return self._pause_state
    
    def play_audio(self,progress_data: GameProgressSchema ):
        keys = pygame.key.get_pressed()
        if keys[pygame.K_m] and (time.time() - self._last_m_key_pressed) > 0.3:
            self.sound_controller.toggle_sound()
            self._last_m_key_pressed = time.time()

        if not isinstance(progress_data, GameProgressSchema):
            progress_data = GameProgressSchema.model_validate(progress_data)
        music_objs = progress_data.musics
        sound_objs = progress_data.sounds
        self.sound_controller.unpause()

        for music_obj in music_objs:
            self.sound_controller.play_music(music_obj['music_id'])
        for sound_obj in sound_objs:
            self.sound_controller.play_sound(sound_obj['sound_id'])

        pass


import Box2D
import pygame

# from .car import Car
from mlgame.game.paia_game import GameResultState
from mlgame.utils.enum import get_ai_name
from mlgame.view.audio_model import SoundProgressSchema
from .env import *
from .gameMode import GameMode
from .maze_wall import Wall
from .points import End_point, Check_point, Outside_point
from .tilemap import Map
from .tiledMap_to_box2d import TiledMap_box2d


class MazeMode(GameMode):
    def __init__(self, user_num: int, map_file, time, sensor):
        super(MazeMode, self).__init__()
        '''load map data'''
        self.user_num = user_num
        # self.maze_id = maze_no - 1
        self.map_file = map_file
        self.load_data()

        '''group of sprites'''
        self.worlds = []
        self.cars = pygame.sprite.Group()
        self.walls = pygame.sprite.Group()
        self.slant_walls = pygame.sprite.Group()
        self.all_points = pygame.sprite.Group()  # Group inclouding end point, check points,etc.

        '''data set'''
        self.wall_info = []
        self.wall_vertices_for_Box2D = []
        self.car_info = []
        self.ranked_user = []  # pygame.sprite car
        self.result = []
        self.eliminated_user = []
        self.sound=[]

        self.game_end_time = time  # int, decide how many frames the game will end even some users don't finish game
        pygame.font.init()
        self.state = GameResultState.UN_PASSED
        self.is_end = False
        self.sensor_num = sensor
        self.x = 0
        self._init_world(user_num)
        self.new()
        # self.world.contactListener.fixtureA = self.car
        # print(self.world.contactListener.m_contactList)

    def new(self):
        # initialize all variables and do all setup for a new game
        self.pygame_point = [10, 100]
        map = TiledMap_box2d(self.map_file, 32)
        walls = map.get_wall_info()
        check_walls = map.get_check_wall_info()
        for wall in walls:
            vertices = map.transfer_to_box2d(wall)
            for world in self.worlds:
                wall = Wall(self, vertices, world)
                if self.worlds.index(world) == 0:
                    self.walls.add(wall)
        for check_wall in check_walls:
            vertices = map.transfer_to_box2d(check_wall)
            vertices = [self.transfer_box2d_to_pygame(v) for v in vertices]
            cp = Check_point(self, vertices)
            self.all_points.add(cp)
            self.check_point_num += 1
            self.check_points.append(cp)
        obj = map.load_other_obj()
        self.load_map_object(obj)
        for wall in self.walls:
            vertices = [(wall.body.transform * v) for v in wall.box.shape.vertices]
            self.wall_info.append([vertices[0], vertices[1]])
            self.wall_info.append([vertices[2], vertices[1]])
            self.wall_info.append([vertices[3], vertices[0]])
            self.wall_info.append([vertices[2], vertices[3]])

        for car in self.cars:
            car.rect.center = self.transfer_box2d_to_pygame(car.body.position)
            car.detect_distance(self.frame, self.wall_info)
            self.car_info.append(car.get_info())

    def update_sprite(self, command):
        '''update the model of game,call this fuction per frame'''
        self.car_info.clear()
        self.sound.clear()
        self.frame += 1
        self.handle_event()
        self._is_game_end()
        self.command = command
        for car in self.cars:
            car.update(command[get_ai_name(car.car_no)])
            car.rect.center = self.transfer_box2d_to_pygame(car.body.position)
            car.detect_distance(self.frame, self.wall_info)
            self.car_info.append(car.get_info())

        self.all_points.update()
        for car in self.cars:
            if len(car.body.contacts) > 2:
                contact = 0
                for contact_edge in car.body.contacts:
                    if contact_edge.contact.touching:
                        contact += 1
                if contact > 2:
                    if car.collide(self.frame):
                        self.sound.append(
                            SoundProgressSchema(sound_id='bomb').__dict__
                        )

                        pass
                        # self.sound_controller.play_bomb_sound()
        for world in self.worlds:
            world.Step(TIME_STEP, 10, 10)
            world.ClearForces()
        if self.is_end:
            self.running = False

    def limit_pygame_screen(self):
        keystate = pygame.key.get_pressed()
        if keystate[pygame.K_w]:
            self.pygame_point[1] += 0.2
        elif keystate[pygame.K_s]:
            self.pygame_point[1] -= 0.2
        elif keystate[pygame.K_a]:
            self.pygame_point[0] -= 0.2
        elif keystate[pygame.K_d]:
            self.pygame_point[0] += 0.2

        if self.pygame_point[1] > 0:
            self.pygame_point[1] = 0
        elif self.pygame_point[1] < TILE_HEIGHT / PPM - self.map.tileHeight:
            self.pygame_point[1] = TILE_HEIGHT / PPM - self.map.tileHeight
        else:
            pass
        if self.pygame_point[0] > self.map.tileWidth - TILE_WIDTH / PPM:
            self.pygame_point[0] = self.map.tileWidth - TILE_WIDTH / PPM
        elif self.pygame_point[0] < 0:
            self.pygame_point[0] = 0
        else:
            pass

    def load_data(self):
        try:
            self.map = Map(self.map_file)
        except Exception:
            print(f"File '{self.map_file}' is not found.We will load first map for you.")
            self.map_file = path.join(MAP_DIR, "level_1.tmj")
            self.map = Map(self.map_file)

    def _init_world(self, user_no: int):
        self.contact_man = Box2D.b2ContactManager()
        for i in range(self.user_num):
            self.worlds.append(Box2D.b2.world(gravity=(0, 0), doSleep=True, CollideConnected=False,
                                              contactListener=self.contact_man.contactListener))
        # self.world = Box2D.b2.world(gravity=(0, 0), doSleep=True, CollideConnected=False, contactListener=self.contact_man.contactListener)

    def _is_game_end(self):
        """
            遊戲結束條件
            1. 全部玩家抵達終點
            2. 時間結束
        """
        if self.frame >= self.game_end_time:
            for car in self.cars:
                if car not in self.eliminated_user and car.is_running:
                    self.eliminated_user.append(car)
                    car.is_running = False
                    car.status = "GAME_OVER"
            self.is_end = True
            self.ranked_user = self.rank()
            self._print_result()
            self.status = "END"

        elif len(self.cars) == len(self.eliminated_user):
            self.is_end = True
            self.ranked_user = self.rank()
            self._print_result()
            self.status = "END"



from os import path

import pygame
from mlgame.view.view_model import create_asset_init_data, create_image_view_data

from .env import IMAGE_DIR, WINDOW_HEIGHT, WINDOW_WIDTH


class Station(pygame.sprite.Sprite):
    def __init__(self, construction, **kwargs):
        super().__init__()
        self.id = construction["_id"]
        self.rect = pygame.Rect(construction["_init_pos"], construction["_init_size"])
        self.power = kwargs["capacity"]
        self.angle = 0
        if self.rect.x >= WINDOW_WIDTH // 2 and self.rect.y < (WINDOW_HEIGHT - 100) // 2:
            self.quadrant = 1
        elif self.rect.x < WINDOW_WIDTH // 2 and self.rect.y < (WINDOW_HEIGHT - 100) // 2:
            self.quadrant = 2
        elif self.rect.x < WINDOW_WIDTH // 2 and self.rect.y >= (WINDOW_HEIGHT - 100) // 2:
            self.quadrant = 3
        else:
            self.quadrant = 4

        self.spawn_cd = kwargs["spawn_cd"]
        self.cooldown = 0
        self.is_alive = True

    def update(self):
        if not self.is_alive:
            self.cooldown -= 1
            if self.cooldown <= 0:
                self.is_alive = True

    def collect(self):
        self.is_alive = False
        self.cooldown = self.spawn_cd

    def get_data_from_obj_to_game(self):
        if 5 == self.id:
            info = {"id": "oil", "x": self.rect.x, "y": self.rect.y, "power": self.power if self.is_alive else 0}
        else:
            info = {"id": "bullets", "x": self.rect.x, "y": self.rect.y, "power": self.power if self.is_alive else 0}
        return info

    def get_obj_progress_data(self):
        if not self.is_alive:
            return []
        if 5 == self.id:
            return create_image_view_data(f"oil", self.rect.x, self.rect.y
                                          , self.rect.width, self.rect.height, self.angle)
        else:
            return create_image_view_data(f"bullets", self.rect.x, self.rect.y
                                          , self.rect.width, self.rect.height, self.angle)

    def get_obj_init_data(self):
        bullets_id = "bullets"
        oil_id = "oil"
        bullets_url = f"https://raw.githubusercontent.com/Jesse-Jumbo/TankMan/main/asset/image/{bullets_id}.svg"
        oil_url = f"https://raw.githubusercontent.com/Jesse-Jumbo/TankMan/main/asset/image/{oil_id}.svg"
        image_init_data = [create_asset_init_data(bullets_id, self.rect.width, self.rect.height,
                                                  path.join(IMAGE_DIR, f"{bullets_id}.png"), bullets_url)
                           , create_asset_init_data(oil_id, self.rect.width, self.rect.height,
                                                    path.join(IMAGE_DIR, f"{oil_id}.png"), oil_url)]
        return image_init_data



import json
import os

from orjson import orjson


def save_json(dest_folder, game_result:dict):
    try:
        with open(os.path.join(dest_folder,"result.json"), "w") as f:
            f.write(orjson.dumps(game_result).decode())
        pass
    except Exception as e:
        print(f"Save result.json in {dest_folder} failed. Game result is : {game_result}")



def check_folder_existed_and_readable_or_create(path):
    if not os.path.exists(path):
        os.makedirs(path)

    if os.path.isdir(path) and os.access(path, os.R_OK):
        return path
    else:
        raise NotADirectoryError(f'{path} is not a readable directory or does not exist')



import sys

import setuptools

# readme.md = github readme.md, 這裡可接受markdown寫法
# 如果沒有的話，需要自己打出介紹此專案的檔案，再讓程式知道
sys.path.append(r".")
import mlgame.version

with open("README.md", "r") as fh:
    long_description = fh.read()

def _read_requirements(file_path: str) -> list:
    requirements = []
    try:
        with open(file_path, "r") as rf:
            for raw_line in rf:
                line = raw_line.strip()
                if not line or line.startswith("#"):
                    continue
                if line.startswith(("-r", "--requirement")):
                    # support nested requirement files: -r other.txt
                    parts = line.split(maxsplit=1)
                    if len(parts) == 2:
                        requirements.extend(_read_requirements(parts[1]))
                    continue
                if line.startswith(("-e", "--editable")):
                    # ignore editable installs
                    continue
                requirements.append(line)
    except FileNotFoundError:
        # 如果發佈包中沒有 requirements.txt，不要阻斷安裝流程
        requirements = []
    return requirements

setuptools.setup(
    name="mlgame",  #
    version=mlgame.version.version,
    author="PAIA-Tech",
    author_email="service@paia-tech.com",
    description="A machine learning game framework based on Pygame",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PAIA-Playful-AI-Arena/MLGame",
    packages=setuptools.find_packages(
        exclude=["tests"]
    ),
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10.0, <3.13.0',
    include_package_data=True,
    keywords=["AI", "machine learning", 'game', 'framework'],

    install_requires=_read_requirements("requirements.txt")

)



import pygame
import time
from .env import *
import random
from mlgame.game.paia_game import GameStatus

class Car(pygame.sprite.Sprite):
    def __init__(self, y,distance):
        pygame.sprite.Sprite.__init__(self)
        self.rect = pygame.Rect(distance+20, y, car_size[0],car_size[1])
        self.state = True
        self.status = None
        self.velocity = 0
        self.distance = distance
        self.car_no = 0
        self.car_info = {}
        self.coin_num = 0
        self.max_vel = random.randint(10, 14)

    def speedUp(self):
        self.velocity += 0.01*(self.velocity**0.6)+0.04

    def brakeDown(self):
        self.velocity -= 0.2

    def slowDown(self):
        if self.velocity > 1:
            self.velocity -= 0.05
        elif 0 <= self.velocity < 0.9:
            self.velocity += 0.3
    def moveRight(self):
        self.rect.centery += 3

    def moveLeft(self):
        self.rect.centery -= 3

    def keep_in_screen(self):
        if self.rect.left < -500 or self.rect.right > 1500 or self.rect.top < 100:
            self.state = False

    def get_info(self):
        self.car_info = {"id": self.car_no,
                         "x":self.rect.x,
                         "y":self.rect.y,
                         "status":self.status,
                         "distance": self.distance,
                         "velocity": self.velocity,
                         "coin_num": self.coin_num}
        return self.car_info

class UserCar(Car):
    def __init__(self, y,distance,user_no):
        Car.__init__(self, y,distance)
        self.car_no = user_no
        self.status = GameStatus.GAME_ALIVE
        self.coin_num = 0
        self.max_vel = 15
        self.cash_frame = 0
        self.used_frame = 0

    def update(self, control_list):
        if self.state:
            self.used_frame += 1
            self.handleKeyEvent(control_list)
            self.distance += self.velocity
            self.keep_in_screen()
        else:
            pass

    def keep_in_screen(self):
        if self.velocity > self.max_vel:
            self.velocity = self.max_vel
        elif self.velocity < 0:
            self.velocity = 0

    def handleKeyEvent(self, control_list: list):
        if control_list == None:
            return True
        if LEFT_cmd in control_list:
            self.moveLeft()
            self.max_vel = 14
        if RIGHT_cmd in control_list:
            self.moveRight()
            self.max_vel = 14

        if LEFT_cmd not in control_list and RIGHT_cmd not in control_list:
            self.max_vel = 15
        if SPEED_cmd in control_list:
            self.speedUp()
        elif BRAKE_cmd in control_list:
            self.brakeDown()
        else:
            self.slowDown()

class ComputerCar(Car):
    def __init__(self, y,distance,x):
        Car.__init__(self,y,distance)
        self.rect.left, self.rect.top = x, y
        self.velocity = random.randint(8,12)
        self.car_no = random.randrange(101, 200)
        self.distance = distance
        self.action = None

    def update(self,cars):
        if self.state:
            for car in cars:
                if abs(self.rect.centery - car.rect.centery) < 40:
                    self.detect_other_cars(car)
                else:
                    continue
                if self.action == "stop":
                    self.brakeDown()
                    break
                else:
                    self.speedUp()
            self.distance += self.velocity
            if self.velocity < 0:
                self.velocity = 0
            if self.velocity > self.max_vel:
                self.velocity = self.max_vel
            pass
        else:
            pass
        self.keep_in_screen()

    def detect_other_cars(self, car):
        distance = car.rect.centerx - self.rect.centerx
        if 200 > distance > 0:
            self.action = "stop"
        else:
            self.action = "continue"

class Camera():
    def __init__(self, length):
        self.position = 500
        self.velocity = 0
        self.length = length

    def update(self,car_velocity):
        self.revise_velocity(car_velocity)
        self.position += self.velocity
        # if self.position > self.length:
        #     self.position = self.length

    def revise_velocity(self,car_velocity):
        if self.position > self.length:
            self.velocity = 0
            return None
        elif self.position > 20000:
            self.velocity = car_velocity+0.1
            return None

        if car_velocity >= 13.5:
            self.velocity = car_velocity-0.5

        elif car_velocity == 0:
            self.velocity = 0.01
        else:
            if self.velocity < car_velocity:
                self.velocity += 0.07
            elif self.velocity > car_velocity+1:
                self.velocity -= 0.07
            else:
                pass



import abc
import pygame


class I_Commander(abc.ABC):
    @abc.abstractmethod
    def getControlDict(self) -> dict:
        pass


keyboardSet = [
    {"TURN_LEFT": pygame.K_LEFT,
     "TURN_RIGHT": pygame.K_RIGHT,
     "SPEED": pygame.K_UP,
     "BRAKE": pygame.K_DOWN},

    {"TURN_LEFT": pygame.K_SPACE,
     "TURN_RIGHT": pygame.K_SPACE,
     "SPEED": pygame.K_SPACE,
     "BRAKE": pygame.K_SPACE},

    {"TURN_LEFT": pygame.K_SPACE,
     "TURN_RIGHT": pygame.K_SPACE,
     "SPEED": pygame.K_SPACE,
     "BRAKE": pygame.K_SPACE},

    {"TURN_LEFT": pygame.K_SPACE,
     "TURN_RIGHT": pygame.K_SPACE,
     "SPEED": pygame.K_SPACE,
     "BRAKE": pygame.K_SPACE},

    {"TURN_LEFT": pygame.K_SPACE,
     "TURN_RIGHT": pygame.K_SPACE,
     "SPEED": pygame.K_SPACE,
     "BRAKE": pygame.K_SPACE},

    {"TURN_LEFT": pygame.K_SPACE,
     "TURN_RIGHT": pygame.K_SPACE,
     "SPEED": pygame.K_SPACE,
     "BRAKE": pygame.K_SPACE}

]


class KeyBoardCommander(I_Commander):
    def __init__(self, keyboard_no=0):
        self.no = keyboard_no
        self.speedKey = keyboardSet[keyboard_no]["SPEED"]
        self.brakeKey = keyboardSet[keyboard_no]["BRAKE"]
        self.moveLeftKey = keyboardSet[keyboard_no]["TURN_LEFT"]
        self.moveRightKey = keyboardSet[keyboard_no]["TURN_RIGHT"]
        self.speed = 100

    def getControlDict(self):
        keys = pygame.key.get_pressed()
        control_list = [{"left_PWM": 0, "right_PWM": 0}]
        control_dic = {"LEFT": keys[self.moveLeftKey],
                       "RIGHT": keys[self.moveRightKey],
                       "SPEED_UP": keys[self.speedKey],
                       "BRAKEDOWN": keys[self.brakeKey]}

        if control_dic["SPEED_UP"]:
            control_list[0]["left_PWM"] = self.speed
            control_list[0]["right_PWM"] = self.speed
        elif control_dic["BRAKEDOWN"]:
            control_list[0]["left_PWM"] = -1 * self.speed
            control_list[0]["right_PWM"] = -1 * self.speed

        if control_dic["LEFT"]:
            control_list[0]["right_PWM"] += self.speed
            # control_list[0]["left_PWM"] -= self.speed
        elif control_dic["RIGHT"]:
            control_list[0]["left_PWM"] += self.speed

        return control_list



import json
from os import path

import orjson
import pydantic
import pytest

from mlgame.argument.tool import revise_ai_clients, get_data_from_json_file
from mlgame.argument.model import UserNumConfig
from mlgame.core.model import GameProgressSchema
from mlgame.view.audio_model import MusicProgressSchema


def test_UserNum():
    try:
        user_num_config = UserNumConfig(min=2, max=1)
        assert False
    except Exception as e:
        assert isinstance(e, pydantic.ValidationError)


@pytest.mark.parametrize(
    "   ai_clients_files        ,min,max,expected_ai_count",
    [
        (['./1p.py', './2p.py'] ,1  ,2  ,2),
        (['./1p.py', './2p.py'] ,1  ,4  ,2),
        (['./1p.py']            ,1  ,2  ,1),
        (['./1p.py']            ,2  ,2  ,2),
        (['./1p.py']            ,4  ,5  ,4),
        (['./1p.py','./2p.py']  ,4  ,5  ,4),
        (['./1p.py','./2p.py']  ,4  ,5  ,4),
        pytest.param(['./1p.py','./2p.py']  ,4  ,5  ,5, marks=pytest.mark.xfail)
    ])
def test_revise_ai_clients(ai_clients_files: list, min: int, max: int, expected_ai_count: int):

    user_num_config = UserNumConfig(min=min, max=max)
    ai_clients = revise_ai_clients(ai_clients_files, user_num_config=user_num_config)
    assert len(ai_clients) == expected_ai_count


def test_open_config_file():
    config_file = path.join(path.dirname(__file__), "test_data", "game_config.json")
    config_data = get_data_from_json_file(config_file)
    assert isinstance(config_data, dict)


def test_dict_contains_key():
    obj = {'key':'value'}

    if 'key' in obj:
        var = obj['key']
    obj = {}
    assert obj.get('key') is None

    music ={'music_id':'musiccc'}
    MusicProgressSchema(**music)

    music = {'type':'music','music_id': 'musiccc'}
    MusicProgressSchema(**music)


def test_orjson_parse_obj():
    data = GameProgressSchema(frame=1)

    assert orjson.dumps(data.model_dump())
    assert json.dumps(data.model_dump())


import Box2D
import pygame
from .env import *


def count_position(vertices):
    v_sum = (0, 0)
    for vertice in vertices:
        v_sum += vertice
    position = v_sum[0] / 4, v_sum[1] / 4
    return position

class Wall(pygame.sprite.Sprite):
    def __init__(self, game, vertices, world):
        pygame.sprite.Sprite.__init__(self)
        self.game = game
        self.world = world
        self.x, self.y = (0, 0)
        self.body = world.CreateKinematicBody(position = (0, 0))
        self.box = self.body.CreatePolygonFixture(box = ((TILESIZE/ (2*PPM), TILESIZE/ (2*PPM))), vertices = vertices)

class VerticalMoveWall(Wall):
    def __init__(self, game, vertices, world, velocity, distance=None):
        Wall.__init__(self, game, vertices, world)
        if distance > 0:
            self.start_coordinate = self.body.position[0], self.body.position[1]
            self.end_coordinate = self.body.position[0], self.body.position[1] + distance
        elif distance < 0:
            self.start_coordinate = self.body.position[0], self.body.position[1] + distance
            self.end_coordinate = self.body.position[0], self.body.position[1]
        self.velocity = abs(velocity)
        self.body.linearVelocity = (0, self.velocity)

    def update(self, *args, **kwargs) -> None:
        if self.body.position[1] > self.end_coordinate[1]:
            self.body.linearVelocity = (0, -self.velocity)
        elif self.body.position[1] < self.start_coordinate[1]:
            self.body.linearVelocity = (0, self.velocity)

class HorizontalMoveWall(Wall):
    def __init__(self, game, vertices, world, velocity, distance):
        Wall.__init__(self, game, vertices, world)
        if distance > 0:
            self.start_coordinate = self.body.position[0], self.body.position[1]
            self.end_coordinate = self.body.position[0] + distance, self.body.position[1]
        elif distance < 0:
            self.start_coordinate = self.body.position[0] + distance, self.body.position[1]
            self.end_coordinate = self.body.position[0], self.body.position[1]

        self.velocity = abs(velocity)
        self.body.linearVelocity = (self.velocity, 0)

    def update(self, *args, **kwargs) -> None:
        if self.body.position[0] > self.end_coordinate[0]:
            self.body.linearVelocity = (-self.velocity, 0)
        elif self.body.position[0] < self.start_coordinate[0]:
            self.body.linearVelocity = (self.velocity, 0)

class SlantWall(Wall):
    def __init__(self, game, vertices, world):
        Wall.__init__(self, game, vertices, world)
        self.game = game
        self.world = world
        self.x, self.y = (0, 0)
        self.body = world.CreateKinematicBody(position = (0, 0))
        self.box = self.body.CreatePolygonFixture(vertices = vertices)
        pass
